"""
Serializers for assessment details for the ESG app
"""

from rest_framework import serializers
from openassessment.assessment.models.base import Assessment, AssessmentPart


# pylint: disable=abstract-method
class SubmissionDetailFileSerilaizer(serializers.Serializer):
    """
    Serialized info about a single file
    """
    download_url = serializers.URLField()
    description = serializers.CharField()
    name = serializers.CharField()
    size = serializers.IntegerField()


class AssessmentPartSerializer(serializers.ModelSerializer):
    name = serializers.CharField(source='criterion.name')
    option = serializers.CharField(source='option.name', default=None)
    points = serializers.IntegerField(source='option.points', default=None)

    class Meta:
        model = AssessmentPart
        fields = ['name', 'option', 'points', 'feedback']
        read_only_fields = fields


class AssessmentScoreSerializer(serializers.ModelSerializer):
    """
    Serializer for pulling score info off the assessment and into a "score" dict
    """
    pointsEarned = serializers.IntegerField(source='points_earned')
    pointsPossible = serializers.IntegerField(source='points_possible')

    class Meta:
        model = Assessment
        fields = [
            'pointsEarned',
            'pointsPossible',
        ]


class AssessmentSerializer(serializers.ModelSerializer):
    """
    Serialized info about an assessment
    """
    criteria = AssessmentPartSerializer(source='parts', many=True, read_only=True)
    score = AssessmentScoreSerializer(source='*')

    class Meta:
        model = Assessment
        fields = [
            'feedback',
            'score',
            'criteria',
        ]
        read_only_fields = fields



"""
Helper class for loading generated file from webpack.
"""
import json
import logging

from pkg_resources import resource_string
from django.conf import settings

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


def urljoin(*args):
    """
    Joining multiple url continuously. The default behavior from os.path.join
    would completely reset the path if the second path start with '/'.
    """
    begining_slash = '/' if args[0].startswith('/') else ''
    trailing_slash = '/' if args[-1].endswith('/') else ''
    joined_path = '/'.join(map(lambda x: str(x).strip('/'), args))
    return begining_slash + joined_path + trailing_slash


class LoadStatic:
    """
    Helper class for loading generated file from webpack.
    """

    _manifest = {}
    _base_url = ''
    _is_loaded = False

    @staticmethod
    def reload_manifest():
        """
        Reload from manifest file
        """
        root_url, base_url = '', '/static/dist/'
        if hasattr(settings, 'LMS_ROOT_URL'):
            root_url = settings.LMS_ROOT_URL
        else:
            logger.error('LMS_ROOT_URL is undefined')

        try:
            json_data = resource_string(__name__, 'static/dist/manifest.json').decode("utf8")
            LoadStatic._manifest = json.loads(json_data)
            LoadStatic._is_loaded = True
        except OSError:
            logger.error('Cannot find static/dist/manifest.json')
        finally:
            LoadStatic._base_url = urljoin(root_url, base_url)

    @staticmethod
    def get_url(key):
        """
        get url from key
        """
        if not LoadStatic._is_loaded:
            LoadStatic.reload_manifest()
        url = LoadStatic._manifest[key] if key in LoadStatic._manifest else key
        return urljoin(LoadStatic._base_url, url)



# Generated by Django 2.2.10 on 2020-03-27 19:25

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('workflow', '0002_remove_django_extensions'),
    ]

    operations = [
        migrations.CreateModel(
            name='TeamAssessmentWorkflow',
            fields=[
                ('assessmentworkflow_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='workflow.AssessmentWorkflow')),
                ('team_submission_uuid', models.CharField(max_length=128, unique=True)),
            ],
            options={
                'abstract': False,
            },
            bases=('workflow.assessmentworkflow',),
        ),
    ]



"""
The Peer Assessment Mixin for all Peer Functionality.

"""

import logging

from webob import Response
from xblock.core import XBlock
from openassessment.assessment.errors import (PeerAssessmentInternalError, PeerAssessmentRequestError,
                                              PeerAssessmentWorkflowError)
from openassessment.workflow.errors import AssessmentWorkflowError
from openassessment.xblock.defaults import DEFAULT_RUBRIC_FEEDBACK_TEXT

from .data_conversion import (clean_criterion_feedback, create_rubric_dict, create_submission_dict,
                              verify_assessment_parameters)
from .resolve_dates import DISTANT_FUTURE
from .user_data import get_user_preferences

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class PeerAssessmentMixin:
    """The Peer Assessment Mixin for all Peer Functionality.

    Abstracts all functionality and handlers associated with Peer Assessment.
    All Peer Assessment API calls should be contained without this Mixin as
    well.

    PeerAssessmentMixin is a Mixin for the OpenAssessmentBlock. Functions in
    the PeerAssessmentMixin call into the OpenAssessmentBlock functions and
    will not work outside of OpenAssessmentBlock

    """

    @XBlock.json_handler
    @verify_assessment_parameters
    def peer_assess(self, data, suffix=''):  # pylint: disable=unused-argument
        """Place a peer assessment into OpenAssessment system

        Assess a Peer Submission.  Performs basic workflow validation to ensure
        that an assessment can be performed as this time.

        Args:
            data (dict): A dictionary containing information required to create
                a new peer assessment.  This dict should have the following attributes:
                `submission_uuid` (string): The unique identifier for the submission being assessed.
                `options_selected` (dict): Dictionary mapping criterion names to option values.
                `overall_feedback` (unicode): Written feedback for the submission as a whole.
                `criterion_feedback` (unicode): Written feedback per the criteria for the submission.

        Returns:
            Dict with keys "success" (bool) indicating success/failure.
            and "msg" (unicode) containing additional information if an error occurs.

        """
        # Import is placed here to avoid model import at project startup.
        from openassessment.assessment.api import peer as peer_api
        if self.submission_uuid is None:
            return {
                'success': False, 'msg': self._('You must submit a response before you can perform a peer assessment.')
            }

        uuid_server, uuid_client = self._get_server_and_client_submission_uuids(data)
        if uuid_server != uuid_client:
            logger.warning(
                'Irrelevant assessment submission: expected "%s", got "%s"',
                uuid_server,
                uuid_client,
            )
            return {
                'success': False,
                'msg': self._('This feedback has already been submitted or the submission has been cancelled.'),
            }

        assessment_ui_model = self.get_assessment_module('peer-assessment')
        if assessment_ui_model:
            try:
                # Create the assessment
                assessment = peer_api.create_assessment(
                    self.submission_uuid,
                    self.get_student_item_dict()["student_id"],
                    data['options_selected'],
                    clean_criterion_feedback(self.rubric_criteria_with_labels, data['criterion_feedback']),
                    data['overall_feedback'],
                    create_rubric_dict(self.prompts, self.rubric_criteria_with_labels),
                    assessment_ui_model['must_be_graded_by']
                )

                # Emit analytics event...
                self.publish_assessment_event("openassessmentblock.peer_assess", assessment)

            except (PeerAssessmentRequestError, PeerAssessmentWorkflowError):
                logger.warning(
                    "Peer API error for submission UUID %s",
                    self.submission_uuid,
                    exc_info=True
                )
                return {'success': False, 'msg': self._("Your peer assessment could not be submitted.")}
            except PeerAssessmentInternalError:
                logger.exception(
                    "Peer API internal error for submission UUID: %s",
                    self.submission_uuid
                )
                msg = self._("Your peer assessment could not be submitted.")
                return {'success': False, 'msg': msg}

            # Update both the workflow that the submission we're assessing
            # belongs to, as well as our own (e.g. have we evaluated enough?)
            try:
                if assessment:
                    self.update_workflow_status(submission_uuid=assessment['submission_uuid'])
                self.update_workflow_status()
            except AssessmentWorkflowError:
                logger.exception(
                    "Workflow error occurred when submitting peer assessment "
                    "for submission %s",
                    self.submission_uuid
                )
                msg = self._('Could not update workflow status.')
                return {'success': False, 'msg': msg}

            # Temp kludge until we fix JSON serialization for datetime
            assessment["scored_at"] = str(assessment["scored_at"])

            return {'success': True, 'msg': ''}

        return {'success': False, 'msg': self._('Could not load peer assessment.')}

    @XBlock.handler
    def render_peer_assessment(self, data, suffix=''):  # pylint: disable=unused-argument
        """Renders the Peer Assessment HTML section of the XBlock

        Generates the peer assessment HTML for the first section of an Open
        Assessment XBlock. See OpenAssessmentBlock.render_assessment() for
        more information on rendering XBlock sections.

        Args:
            data (dict): May contain an attribute 'continue_grading', which
                allows a student to continue grading peers past the required
                number of assessments.

        """
        if "peer-assessment" not in self.assessment_steps:
            return Response("")
        continue_grading = data.params.get('continue_grading', False)
        path, context_dict = self.peer_path_and_context(continue_grading)

        # For backwards compatibility, if no feedback default text has been
        # set, use the default text
        if 'rubric_feedback_default_text' not in context_dict:
            context_dict['rubric_feedback_default_text'] = DEFAULT_RUBRIC_FEEDBACK_TEXT

        return self.render_assessment(path, context_dict)

    def peer_path_and_context(self, continue_grading):
        """
        Return the template path and context for rendering the peer assessment step.

        Args:
            continue_grading (bool): If true, the user has chosen to continue grading.

        Returns:
            tuple of (template_path, context_dict)

        """
        # Import is placed here to avoid model import at project startup.
        from openassessment.assessment.api import peer as peer_api
        path = 'openassessmentblock/peer/oa_peer_unavailable.html'
        finished = False
        problem_closed, reason, start_date, due_date = self.is_closed(step="peer-assessment")
        user_preferences = get_user_preferences(self.runtime.service(self, 'user'))

        context_dict = {
            "rubric_criteria": self.rubric_criteria_with_labels,
            "allow_multiple_files": self.allow_multiple_files,
            "allow_latex": self.allow_latex,
            "prompts_type": self.prompts_type,
            "user_timezone": user_preferences['user_timezone'],
            "user_language": user_preferences['user_language'],
            "xblock_id": self.get_xblock_id(),
        }

        if self.rubric_feedback_prompt is not None:
            context_dict["rubric_feedback_prompt"] = self.rubric_feedback_prompt

        if self.rubric_feedback_default_text is not None:
            context_dict['rubric_feedback_default_text'] = self.rubric_feedback_default_text

        # We display the due date whether the problem is open or closed.
        # If no date is set, it defaults to the distant future, in which
        # case we don't display the date.
        if due_date < DISTANT_FUTURE:
            context_dict['peer_due'] = due_date

        workflow = self.get_workflow_info()
        workflow_status = workflow.get('status')
        peer_complete = workflow.get('status_details', {}).get('peer', {}).get('complete', False)
        peer_skipped = workflow.get('status_details', {}).get('peer', {}).get('skipped', False)
        continue_grading = continue_grading and peer_complete

        student_item = self.get_student_item_dict()
        assessment = self.get_assessment_module('peer-assessment')
        if assessment:
            context_dict["must_grade"] = assessment["must_grade"]
            finished, count = peer_api.has_finished_required_evaluating(
                self.submission_uuid,
                assessment["must_grade"]
            )
            context_dict["graded"] = count
            context_dict["review_num"] = count + 1

            if continue_grading:
                context_dict["submit_button_text"] = self._(
                    "Submit your assessment and review another response"
                )
            elif assessment["must_grade"] - count == 1:
                context_dict["submit_button_text"] = self._(
                    "Submit your assessment and move to next step"
                )
            else:
                context_dict["submit_button_text"] = self._(
                    "Submit your assessment and move to response #{response_number}"
                ).format(response_number=(count + 2))

        if workflow_status == "cancelled":
            path = 'openassessmentblock/peer/oa_peer_cancelled.html'
            # Sets the XBlock boolean to signal to Message that it WAS able to grab a submission
            self.no_peers = True

        # Once a student has completed a problem, it stays complete,
        # so this condition needs to be first.
        elif (workflow.get('status') == 'done' or finished) and not continue_grading:
            path = "openassessmentblock/peer/oa_peer_complete.html"

        # Allow continued grading even if the problem due date has passed
        elif continue_grading and student_item:
            peer_sub = self.get_peer_submission(student_item, assessment)
            if peer_sub:
                path = 'openassessmentblock/peer/oa_peer_turbo_mode.html'
                context_dict["peer_submission"] = create_submission_dict(peer_sub, self.prompts)

                # Determine if file upload is supported for this XBlock.
                context_dict["file_upload_type"] = self.file_upload_type
                context_dict["peer_file_urls"] = self.get_download_urls_from_submission(peer_sub)
            else:
                path = 'openassessmentblock/peer/oa_peer_turbo_mode_waiting.html'
        elif reason == 'due' and problem_closed:
            path = 'openassessmentblock/peer/oa_peer_closed.html'
        elif reason == 'start' and problem_closed:
            context_dict["peer_start"] = start_date
            path = 'openassessmentblock/peer/oa_peer_unavailable.html'
        elif workflow.get("status") == "peer" or peer_skipped:
            peer_sub = self.get_peer_submission(student_item, assessment)
            if peer_sub:
                path = 'openassessmentblock/peer/oa_peer_assessment.html'
                context_dict["peer_submission"] = create_submission_dict(peer_sub, self.prompts)
                # Determine if file upload is supported for this XBlock.
                context_dict["file_upload_type"] = self.file_upload_type
                context_dict["peer_file_urls"] = self.get_download_urls_from_submission(peer_sub)
                # Sets the XBlock boolean to signal to Message that it WAS NOT able to grab a submission
                self.no_peers = False
            else:
                path = 'openassessmentblock/peer/oa_peer_waiting.html'
                # Sets the XBlock boolean to signal to Message that it WAS able to grab a submission
                self.no_peers = True

        return path, context_dict

    def get_peer_submission(self, student_item_dict, assessment):
        """
        Retrieve a submission to peer-assess.

        Args:
            student_item_dict (dict): The student item for the student creating the submission.
            assessment (dict): A dict describing the requirements for grading.

        Returns:
            dict: The serialized submission model.

        """
        # Import is placed here to avoid model import at project startup.
        from openassessment.assessment.api import peer as peer_api
        peer_submission = False
        try:
            peer_submission = peer_api.get_submission_to_assess(
                self.submission_uuid,
                assessment["must_be_graded_by"]
            )
            self.runtime.publish(
                self,
                "openassessmentblock.get_peer_submission",
                {
                    "requesting_student_id": student_item_dict["student_id"],
                    "course_id": student_item_dict["course_id"],
                    "item_id": student_item_dict["item_id"],
                    "submission_returned_uuid": (
                        peer_submission["uuid"] if peer_submission else None
                    )
                }
            )
        except PeerAssessmentWorkflowError as err:
            logger.exception(err)

        return peer_submission

    def _get_server_and_client_submission_uuids(self, data={}):  # pylint: disable=dangerous-default-value
        """
        Retrieve the server and client submission_uuids

        Args:
            data (dict): A dictionary containing new peer assessment data
                This dict should have the following attributes:
                - `submission_uuid` (string): Unique identifier for the submission being assessed
                `- options_selected` (dict): Map criterion names to option values
                `- feedback` (unicode): Written feedback for the submission

        Returns:
            tuple: (uuid_server, uuid_client)
        """
        student_item = self.get_student_item_dict()
        assessment = self.get_assessment_module('peer-assessment')
        submission = self.get_peer_submission(student_item, assessment) or {}
        uuid_server = submission.get('uuid', None)
        uuid_client = data.get('submission_uuid', None)
        return uuid_server, uuid_client



"""
Serializers for submission locks
"""
from rest_framework import serializers
from openassessment.staffgrader.models.submission_lock import SubmissionGradingLock


class SubmissionLockSerializer(serializers.ModelSerializer):
    """
    Serialized info about a submission lock.

    Notes:
    - Passing an empty dict returns ony the locked status, useful for frontend which needs to know locked status.
    - Need to provide context={'user_id': <anon-user-id>} for proper "lock_status" serialization.
    """
    requires_context = True

    submission_uuid = serializers.CharField(required=False)
    owner_id = serializers.CharField(required=False)
    created_at = serializers.DateTimeField(required=False)
    lock_status = serializers.SerializerMethodField()

    def get_lock_status(self, instance):
        """
        Get the lock status, one of:
        - "in-progress" - querying user has the active lock
        - "locked" - a lock is owned by another user
        - "unlocked" - no lock exists for this submission

        NOTE: context={'user_id': <anon-user-id>} must be provided to serializer fo this to be accurate
        """
        if not instance or (not instance.is_active):
            return "unlocked"
        elif instance.owner_id == self.context.get('user_id'):
            return "in-progress"
        else:
            return "locked"

    class Meta:
        model = SubmissionGradingLock
        fields = [
            'submission_uuid',
            'owner_id',
            'created_at',
            'lock_status',
        ]
        read_only_fields = fields



""" A mixin for self Assessments. """


import logging

from webob import Response
from xblock.core import XBlock

from openassessment.assessment.api import self as self_api
from openassessment.workflow import api as workflow_api

from .data_conversion import (clean_criterion_feedback, create_rubric_dict, create_submission_dict,
                              verify_assessment_parameters)
from .resolve_dates import DISTANT_FUTURE
from .user_data import get_user_preferences

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class SelfAssessmentMixin:
    """The Self Assessment Mixin for all Self Assessment Functionality.

    Abstracts all functionality and handlers associated with Self Assessment.
    All Self Assessment API calls should be contained within this Mixin as
    well.

    SelfAssessmentMixin is a Mixin for the OpenAssessmentBlock. Functions in
    the SelfAssessmentMixin call into the OpenAssessmentBlock functions and
    will not work outside of OpenAssessmentBlock.
    """

    @XBlock.handler
    def render_self_assessment(self, data, suffix=''):  # pylint: disable=unused-argument
        if "self-assessment" not in self.assessment_steps:
            return Response("")

        try:
            path, context = self.self_path_and_context()
        except Exception:  # pylint: disable=broad-except
            msg = f"Could not retrieve self assessment for submission {self.submission_uuid}"
            logger.exception(msg)
            return self.render_error(self._("An unexpected error occurred."))
        else:
            return self.render_assessment(path, context)

    def self_path_and_context(self):
        """
        Determine the template path and context to use when rendering the self-assessment step.

        Returns:
            tuple of `(path, context)`, where `path` (str) is the path to the template,
            and `context` (dict) is the template context.

        Raises:
            SubmissionError: Error occurred while retrieving the current submission.
            SelfAssessmentRequestError: Error occurred while checking if we had a self-assessment.
        """
        # Import is placed here to avoid model import at project startup.
        from submissions import api as submission_api

        path = 'openassessmentblock/self/oa_self_unavailable.html'
        problem_closed, reason, start_date, due_date = self.is_closed(step="self-assessment")
        user_preferences = get_user_preferences(self.runtime.service(self, 'user'))

        context = {
            'allow_multiple_files': self.allow_multiple_files,
            'allow_latex': self.allow_latex,
            'prompts_type': self.prompts_type,
            "xblock_id": self.get_xblock_id(),
            'user_timezone': user_preferences['user_timezone'],
            'user_language': user_preferences['user_language']
        }

        # We display the due date whether the problem is open or closed.
        # If no date is set, it defaults to the distant future, in which
        # case we don't display the date.
        if due_date < DISTANT_FUTURE:
            context['self_due'] = due_date

        # If we haven't submitted yet, `workflow` will be an empty dict,
        # and `workflow_status` will be None.
        workflow = self.get_workflow_info()
        workflow_status = workflow.get('status')
        self_complete = workflow.get('status_details', {}).get('self', {}).get('complete', False)
        if workflow_status == 'cancelled':
            path = 'openassessmentblock/self/oa_self_cancelled.html'
            # Sets the XBlock boolean to signal to Message that it WAS able to grab a submission
            self.no_peers = True

        elif self_complete:
            path = 'openassessmentblock/self/oa_self_complete.html'
        elif workflow_status == 'self' or problem_closed:
            assessment = self_api.get_assessment(workflow.get("submission_uuid"))

            if assessment is not None:
                path = 'openassessmentblock/self/oa_self_complete.html'
            elif problem_closed:
                if reason == 'start':
                    context["self_start"] = start_date
                    path = 'openassessmentblock/self/oa_self_unavailable.html'
                elif reason == 'due':
                    path = 'openassessmentblock/self/oa_self_closed.html'
            else:
                submission = submission_api.get_submission(self.submission_uuid)
                context["rubric_criteria"] = self.rubric_criteria_with_labels
                context["self_submission"] = create_submission_dict(submission, self.prompts)
                if self.rubric_feedback_prompt is not None:
                    context["rubric_feedback_prompt"] = self.rubric_feedback_prompt

                if self.rubric_feedback_default_text is not None:
                    context['rubric_feedback_default_text'] = self.rubric_feedback_default_text

                # Determine if file upload is supported for this XBlock and what kind of files can be uploaded.
                context["file_upload_type"] = self.file_upload_type
                context['self_file_urls'] = self.get_download_urls_from_submission(submission)

                path = 'openassessmentblock/self/oa_self_assessment.html'
        else:
            # No submission yet or in peer assessment
            path = 'openassessmentblock/self/oa_self_unavailable.html'

        return path, context

    @XBlock.json_handler
    @verify_assessment_parameters
    def self_assess(self, data, suffix=''):  # pylint: disable=unused-argument
        """
        Create a self-assessment for a submission.

        Args:
            data (dict): Must have the following keys:
                options_selected (dict): Dictionary mapping criterion names to option values.

        Returns:
            Dict with keys "success" (bool) indicating success/failure
            and "msg" (unicode) containing additional information if an error occurs.
        """

        if self.submission_uuid is None:
            return {
                'success': False,
                'msg': self._("You must submit a response before you can perform a self-assessment.")
            }

        try:
            assessment = self_api.create_assessment(
                self.submission_uuid,
                self.get_student_item_dict()['student_id'],
                data['options_selected'],
                clean_criterion_feedback(self.rubric_criteria, data['criterion_feedback']),
                data['overall_feedback'],
                create_rubric_dict(self.prompts, self.rubric_criteria_with_labels)
            )
            self.publish_assessment_event("openassessmentblock.self_assess", assessment)

            # After we've created the self-assessment, we need to update the workflow.
            self.update_workflow_status()
        except (self_api.SelfAssessmentRequestError, workflow_api.AssessmentWorkflowRequestError):
            logger.warning(
                "An error occurred while submitting a self assessment "
                "for the submission %s",
                self.submission_uuid,
                exc_info=True
            )
            msg = self._("Your self assessment could not be submitted.")
            return {'success': False, 'msg': msg}
        except (self_api.SelfAssessmentInternalError, workflow_api.AssessmentWorkflowInternalError):
            logger.exception(
                "An error occurred while submitting a self assessment "
                "for the submission %s",
                self.submission_uuid,
            )
            msg = self._("Your self assessment could not be submitted.")
            return {'success': False, 'msg': msg}
        else:
            return {'success': True, 'msg': ""}



"""
Serializers for the training assessment type.
"""


from django.core.cache import cache
from django.db import IntegrityError, transaction

from openassessment.assessment.data_conversion import update_training_example_answer_format
from openassessment.assessment.models import TrainingExample

from .base import RubricSerializer, rubric_from_dict


class InvalidTrainingExample(Exception):
    """
    The training example could not be deserialized.
    """


def validate_training_example_format(example):
    """
    Check whether the serialized training example dict
    has the correct structure.

    Args:
        example (dict): The serialized training example.

    Returns:
        tuple of (is_valid, errors), where `is_valid` is a bool
        and `errors` is a list of error messages.

    """
    errors = []

    if not isinstance(example, dict):
        errors.append("Training example must be a dictionary")

    if 'answer' not in example:
        errors.append('Training example must contain an "answer" field.')

    if 'options_selected' not in example:
        errors.append('Training example must contain an "options_selected" field.')

    is_valid = (len(errors) == 0)
    return is_valid, errors


def serialize_training_example(example):
    """
    Serialize a training example to a dictionary.

    Args:
        example (TrainingExample): The training example to serialize.

    Returns:
        dict

    """
    # Since training examples are immutable, we can safely cache them
    cache_key = example.cache_key_serialized()
    example_dict = cache.get(cache_key)
    if example_dict is None:
        example_dict = {
            'answer': update_training_example_answer_format(example.answer),
            'options_selected': example.options_selected_dict,
            'rubric': RubricSerializer.serialized_from_cache(example.rubric),
        }
        cache.set(cache_key, example_dict)
    return example_dict


@transaction.atomic
def deserialize_training_examples(examples, rubric_dict):
    """
    Deserialize training examples to Django models.

    Args:
        examples (list of dict): The serialized training examples.
        rubric_dict (dict): The serialized rubric.

    Returns:
        list of TrainingExamples

    Raises:
        InvalidRubric
        InvalidRubricSelection
        InvalidTrainingExample

    Example usage:

        >>> options = [
        >>>     {
        >>>         "order_num": 0,
        >>>         "name": "poor",
        >>>         "explanation": "Poor job!",
        >>>         "points": 0,
        >>>     },
        >>>     {
        >>>         "order_num": 1,
        >>>         "name": "good",
        >>>         "explanation": "Good job!",
        >>>         "points": 1,
        >>>     },
        >>>     {
        >>>         "order_num": 2,
        >>>         "name": "excellent",
        >>>         "explanation": "Excellent job!",
        >>>         "points": 2,
        >>>     },
        >>> ]
        >>>
        >>> rubric = {
        >>>     "prompts": [
        >>>         {"description": "Prompt 1"}
        >>>         {"description": "Prompt 2"}
        >>>         {"description": "Prompt 3"}
        >>>     ],
        >>>     "criteria": [
        >>>         {
        >>>             "order_num": 0,
        >>>             "name": "vocabulary",
        >>>             "prompt": "How varied is the vocabulary?",
        >>>             "options": options
        >>>         },
        >>>         {
        >>>             "order_num": 1,
        >>>             "name": "grammar",
        >>>             "prompt": "How correct is the grammar?",
        >>>             "options": options
        >>>         }
        >>>     ]
        >>> }
        >>>
        >>> examples = [
        >>>     {
        >>>         'answer': {
        >>>             'parts': {
        >>>                 [
        >>>                     {'text:' 'Answer part 1'},
        >>>                     {'text:' 'Answer part 2'},
        >>>                     {'text:' 'Answer part 3'}
        >>>                 ]
        >>>             }
        >>>         },
        >>>         'options_selected': {
        >>>             'vocabulary': 'good',
        >>>             'grammar': 'excellent'
        >>>         }
        >>>     },
        >>>     {
        >>>         'answer': u'Doler',
        >>>         'options_selected': {
        >>>             'vocabulary': 'good',
        >>>             'grammar': 'poor'
        >>>         }
        >>>     }
        >>> ]
        >>>
        >>> examples = deserialize_training_examples(examples, rubric)

    """
    # Parse the rubric
    # This will raise an exception if the serialized rubric is invalid.
    rubric = rubric_from_dict(rubric_dict)

    # Parse each example
    created_examples = []
    for example_dict in examples:
        # Try to retrieve the example from the cache
        cache_key, content_hash = TrainingExample.cache_key(
            example_dict['answer'],
            example_dict['options_selected'],
            rubric
        )
        example = cache.get(cache_key)

        # If we couldn't retrieve the example from the cache, create it
        if example is None:
            # Validate the training example
            is_valid, errors = validate_training_example_format(example_dict)
            if not is_valid:
                raise InvalidTrainingExample("; ".join(errors))

            # Get or create the training example
            try:
                example = TrainingExample.objects.get(content_hash=content_hash)
            except TrainingExample.DoesNotExist:
                try:
                    with transaction.atomic():
                        example = TrainingExample.create_example(
                            example_dict['answer'], example_dict['options_selected'], rubric
                        )
                except IntegrityError:
                    example = TrainingExample.objects.get(content_hash=content_hash)

            # Add the example to the cache
            cache.set(cache_key, example)

        created_examples.append(example)

    return created_examples



"""
Errors defined by the workflow API.
"""


import copy


class AssessmentWorkflowError(Exception):
    """An error that occurs during workflow actions.

    This error is raised when the Workflow API cannot perform a requested
    action.

    """


class AssessmentWorkflowInternalError(AssessmentWorkflowError):
    """An error internal to the Workflow API has occurred.

    This error is raised when an error occurs that is not caused by incorrect
    use of the API, but rather internal implementation of the underlying
    services.

    """


class AssessmentWorkflowRequestError(AssessmentWorkflowError):
    """This error is raised when there was a request-specific error

    This error is reserved for problems specific to the use of the API.

    """

    def __init__(self, field_errors):  # pylint: disable=super-init-not-called
        Exception.__init__(self, repr(field_errors))  # pylint: disable=non-parent-init-called
        self.field_errors = copy.deepcopy(field_errors)


class AssessmentWorkflowNotFoundError(AssessmentWorkflowError):
    """This error is raised when no submission is found for the request.

    If a state is specified in a call to the API that results in no matching
    Submissions, this error may be raised.

    """


class AssessmentApiLoadError(AssessmentWorkflowInternalError):
    """
    The assessment API could not be loaded.
    """
    def __init__(self, assessment_name, api_path):
        msg = "Could not load assessment API for {} from {}".format(
            assessment_name, api_path
        )
        super().__init__(msg)



"""
Create dummy submissions and assessments for testing.
"""


import copy
from uuid import uuid4

from django.core.management.base import BaseCommand, CommandError

import loremipsum
from submissions import api as sub_api
from openassessment.assessment.api import peer as peer_api
from openassessment.assessment.api import self as self_api
from openassessment.workflow import api as workflow_api

STEPS = ['peer', 'self']


class Command(BaseCommand):
    """
    Create dummy submissions and assessments for testing.
    This will generate fake (lorem ipsum) data for:
        * Submission response text
        * Assessment rubric definition
        * Assessment rubric scores
        * Assessment feedback
    """

    help = 'Create dummy submissions and assessments'
    args = '<COURSE_ID> <ITEM_ID> <NUM_SUBMISSIONS> <PERCENTAGE>'

    # Number of peer assessments to create per submission
    NUM_PEER_ASSESSMENTS = 3

    # Number of criteria / options in each rubric
    NUM_CRITERIA = 5
    NUM_OPTIONS = 5

    def __init__(self, *args, **kwargs):
        self.self_assessment_required = kwargs.get('self_assessment_required', False)
        kwargs = {}
        super().__init__(*args, **kwargs)
        self._student_items = []

    def handle(self, *args, **options):
        """
        Execute the command.

        Args:
            course_id (unicode): The ID of the course to create submissions for.
            item_id (unicode): The ID of the item in the course to create submissions for.
            num_submissions (int): Number of submissions to create.
            percentage (int or float): Percentage for assessments to be made against submissions.
        """
        if len(args) < 4:
            raise CommandError('Usage: create_oa_submissions <COURSE_ID> <ITEM_ID> <NUM_SUBMISSIONS> <PERCENTAGE>')

        course_id = str(args[0])
        item_id = str(args[1])

        try:
            num_submissions = int(args[2])
        except ValueError as ex:
            raise CommandError('Number of submissions must be an integer') from ex

        try:
            percentage = float(args[3])
            assessments_to_create = (percentage / 100) * num_submissions
        except ValueError as ex:
            raise CommandError('Percentage for completed submissions must be an integer or float') from ex

        print("Creating {num} submissions for {item} in {course}".format(
            num=num_submissions, item=item_id, course=course_id
        ))

        assessments_created = 0

        for sub_num in range(num_submissions):

            print(f"Creating submission {sub_num}")

            # Create a dummy submission
            student_item = {
                'student_id': uuid4().hex[0:10],
                'course_id': course_id,
                'item_id': item_id,
                'item_type': 'openassessment'
            }
            submission_uuid = self._create_dummy_submission(student_item)
            self._student_items.append(student_item)

            # Create a dummy rubric
            rubric, options_selected = self._dummy_rubric()

            # Create peer assessments
            for num in range(self.NUM_PEER_ASSESSMENTS):
                print(f"-- Creating peer-workflow {num}")

                scorer_id = f'test_{num}'

                # The scorer needs to make a submission before assessing
                scorer_student_item = copy.copy(student_item)
                scorer_student_item['student_id'] = scorer_id
                scorer_submission_uuid = self._create_dummy_submission(scorer_student_item)

                # Retrieve the submission we want to score
                # Note that we are NOT using the priority queue here, since we know
                # exactly which submission we want to score.
                peer_api.create_peer_workflow_item(scorer_submission_uuid, submission_uuid)
                if assessments_created < assessments_to_create:
                    print(f"-- Creating peer-assessment {num}")
                    # Create the peer assessment
                    peer_api.create_assessment(
                        scorer_submission_uuid,
                        scorer_id,
                        options_selected, {}, "  ".join(loremipsum.get_paragraphs(2)),
                        rubric,
                        self.NUM_PEER_ASSESSMENTS
                    )
            assessments_created += 1

            if self.self_assessment_required:
                # Create a self-assessment
                print("-- Creating self assessment")
                self_api.create_assessment(
                    submission_uuid, student_item['student_id'],
                    options_selected, {}, "  ".join(loremipsum.get_paragraphs(2)), rubric
                )
        print(f"{assessments_created} assessments being completed for {num_submissions} submissions")

    @property
    def student_items(self):
        """
        Return the list of student items created by the command.
        This is used for testing the command.

        Returns:
            list of serialized StudentItem models
        """
        return self._student_items

    def _create_dummy_submission(self, student_item):
        """
        Create a dummy submission for a student.

        Args:
            student_item (dict): Serialized StudentItem model.

        Returns:
            str: submission UUID
        """
        answer = {'text': "  ".join(loremipsum.get_paragraphs(5))}
        submission = sub_api.create_submission(student_item, answer)
        workflow_api.create_workflow(submission['uuid'], STEPS)
        workflow_api.update_from_assessments(
            submission['uuid'], {'peer': {'must_grade': 1, 'must_be_graded_by': 1}}
        )
        return submission['uuid']

    def _dummy_rubric(self):
        """
        Randomly generate a rubric and select options from it.

        Returns:
            rubric (dict)
            options_selected (dict)
        """
        rubric = {'criteria': []}
        options_selected = {}
        words = loremipsum.Generator().words

        for criteria_num in range(self.NUM_CRITERIA):
            criterion = {
                'name': words[criteria_num],
                'prompt': "  ".join(loremipsum.get_sentences(2)),
                'order_num': criteria_num,
                'options': []
            }

            for option_num in range(self.NUM_OPTIONS):
                criterion['options'].append({
                    'order_num': option_num,
                    'points': option_num,
                    'name': words[option_num],
                    'explanation': "  ".join(loremipsum.get_sentences(2))
                })

            rubric['criteria'].append(criterion)
            options_selected[criterion['name']] = criterion['options'][0]['name']

        return rubric, options_selected



"""
Leaderboard step in the OpenAssessment XBlock.
"""


import logging

from xblock.core import XBlock

from django.utils.translation import gettext as _

from openassessment.assessment.errors import PeerAssessmentError, SelfAssessmentError
from openassessment.data import OraSubmissionAnswerFactory
from openassessment.fileupload import api as file_upload_api
from openassessment.fileupload.exceptions import FileUploadError
from openassessment.xblock.data_conversion import create_submission_dict

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class LeaderboardMixin:
    """Leaderboard Mixin introduces all handlers for displaying the leaderboard

    Abstracts all functionality and handlers associated with the Leaderboard.

    Leaderboard is a Mixin for the OpenAssessmentBlock. Functions in the
    Leaderboard call into the OpenAssessmentBlock functions and will not work
    outside of OpenAssessmentBlock.

    """

    @XBlock.handler
    def render_leaderboard(self, data, suffix=''):  # pylint: disable=unused-argument
        """
        Render the leaderboard.

        Args:
            data: Not used.

        Kwargs:
            suffix: Not used.

        Returns:
            unicode: HTML content of the leaderboard.
        """
        # Import is placed here to avoid model import at project startup.
        from submissions import api as sub_api

        # Retrieve the status of the workflow.  If no workflows have been
        # started this will be an empty dict, so status will be None.
        workflow = self.get_workflow_info()
        status = workflow.get('status')

        # Render the grading section based on the status of the workflow
        try:
            if status == "done":
                path, context = self.render_leaderboard_complete(self.get_student_item_dict())
            else:  # status is 'self' or 'peer', which implies that the workflow is incomplete
                path, context = self.render_leaderboard_incomplete()
        except (sub_api.SubmissionError, PeerAssessmentError, SelfAssessmentError):
            return self.render_error(_("An unexpected error occurred."))
        else:
            return self.render_assessment(path, context)

    def render_leaderboard_complete(self, student_item_dict):
        """
        Render the leaderboard complete state.

        Args:
            student_item_dict (dict): The student item

        Returns:
            template_path (string), tuple of context (dict)
        """
        # Import is placed here to avoid model import at project startup.
        from submissions import api as sub_api

        # Retrieve top scores from the submissions API
        # Since this uses the read-replica and caches the results,
        # there will be some delay in the request latency.
        scores = sub_api.get_top_submissions(
            student_item_dict['course_id'],
            student_item_dict['item_id'],
            student_item_dict['item_type'],
            self.leaderboard_show
        )
        for score in scores:
            raw_score_content_answer = score['content']
            answer = OraSubmissionAnswerFactory.parse_submission_raw_answer(raw_score_content_answer)
            score['files'] = []
            for uploaded_file in answer.get_file_uploads(missing_blank=True):
                file_download_url = self._get_file_download_url(uploaded_file.key)
                if file_download_url:
                    score['files'].append(
                        file_upload_api.FileDescriptor(
                            download_url=file_download_url,
                            description=uploaded_file.description,
                            name=uploaded_file.name,
                            size=uploaded_file.size,
                            show_delete_button=False
                        )._asdict()
                    )
            if 'text' in score['content'] or 'parts' in score['content']:
                submission = {'answer': score.pop('content')}
                score['submission'] = create_submission_dict(submission, self.prompts)
            elif isinstance(score['content'], str):
                pass
            # Currently, we do not handle non-text submissions.
            else:
                score['submission'] = ""

            score.pop('content', None)

        context = {'topscores': scores,
                   'allow_multiple_files': self.allow_multiple_files,
                   'allow_latex': self.allow_latex,
                   'prompts_type': self.prompts_type,
                   'file_upload_type': self.file_upload_type,
                   'xblock_id': self.get_xblock_id()}

        return 'openassessmentblock/leaderboard/oa_leaderboard_show.html', context

    def render_leaderboard_incomplete(self):
        """
        Render the grade incomplete state.

        Returns:
            template_path (string), tuple of context (dict)
        """
        return 'openassessmentblock/leaderboard/oa_leaderboard_waiting.html', {'xblock_id': self.get_xblock_id()}

    def _get_file_download_url(self, file_key):
        """
        Internal function for retrieving the download url at which the file that corresponds
        to the file_key can be downloaded.

        Arguments:
            file_key (string): Corresponding file key.
        Returns:
            file_download_url (string) or empty string in case of error.
        """
        try:
            file_download_url = file_upload_api.get_download_url(file_key)
        except FileUploadError as exc:
            logger.exception(
                'FileUploadError: URL retrieval failed for key %s with error %s',
                file_key=file_key,
                error=exc
            )
            file_download_url = ''
        return file_download_url



"""
Models for locking Submissions for exclusive grading.
Part of Enhanced Staff Grader (ESG).
"""
from django.db import models
from django.utils.timezone import now

from openassessment.assessment.models.staff import StaffWorkflow
from openassessment.staffgrader.errors.submission_lock import SubmissionLockContestedError


class SubmissionGradingLock(models.Model):
    """
    Internal model for locking a submission for exclusive grading
    """
    TIMEOUT = StaffWorkflow.TIME_LIMIT

    # NOTE - submission_uuid can refer to either the team or individual submission
    submission_uuid = models.CharField(max_length=128, db_index=True, unique=True)
    owner_id = models.CharField(max_length=40)
    created_at = models.DateTimeField(default=now)

    class Meta:
        app_label = "staffgrader"
        ordering = ["created_at", "id"]

    @property
    def is_active(self):
        """
        Check if lock is still active (has not timed out)
        """
        return now() < self.created_at + self.TIMEOUT

    @classmethod
    def currently_active(cls):
        """
        Returns a SubmissionGradingLock queryset filtered to only entries that are currently
        'active' (that were created less than SubmissionGradingLock.TIMEOUT ago)
        """
        timeout_threshold = now() - cls.TIMEOUT
        return cls.objects.filter(created_at__gt=timeout_threshold)

    @classmethod
    def get_submission_lock(cls, submission_uuid):
        """
        Get info about a submission grading lock

        Returns: SubmissionGradingLock info or None
        """
        return cls.objects.filter(submission_uuid=submission_uuid).first()

    @classmethod
    def claim_submission_lock(cls, submission_uuid, user_id):
        """
        Try to claim a submission grading lock

        Returns: SubmissionGradingLock or raises Error
        """
        # See if there's already a lock
        current_lock = cls.get_submission_lock(submission_uuid)

        # If there's already an active lock, raise an error
        # Unless the lock owner is trying to reacquire a lock, which is allowed
        if current_lock and current_lock.is_active and current_lock.owner_id != user_id:
            raise SubmissionLockContestedError

        # Otherwise, delete the lock. This is needed so we don't violate the unique submission_uuid constraint
        if current_lock:
            current_lock.delete()

        # Create a new lock
        new_lock = cls.objects.create(
            owner_id=user_id,
            submission_uuid=submission_uuid,
        )

        return new_lock

    @classmethod
    def clear_submission_lock(cls, submission_uuid, user_id):
        """
        Clear an existing lock. Locks can only be cleared by the lock owner

        Returns: None or raises Error
        """
        # Get the current lock, assuming it exists
        current_lock = cls.get_submission_lock(submission_uuid)
        if not current_lock:
            return

        # Only the owner can clear the lock
        if current_lock.owner_id != user_id:
            raise SubmissionLockContestedError

        current_lock.delete()



"""
Public interface for self-assessment.
"""


import logging

from django.db import DatabaseError, transaction

from submissions.api import SubmissionNotFoundError, get_submission_and_student
from openassessment.assessment.errors import SelfAssessmentInternalError, SelfAssessmentRequestError
from openassessment.assessment.models import Assessment, AssessmentPart, InvalidRubricSelection
from openassessment.assessment.serializers import (InvalidRubric, full_assessment_dict, rubric_from_dict,
                                                   serialize_assessments)
from openassessment.assessment.score_type_constants import SELF_TYPE

logger = logging.getLogger("openassessment.assessment.api.self")  # pylint: disable=invalid-name


def submitter_is_finished(submission_uuid, self_requirements):  # pylint: disable=unused-argument
    """
    Check whether a self-assessment has been completed for a submission.

    Args:
        submission_uuid (str): The unique identifier of the submission.
        self_requirements (dict): Any attributes of the assessment module required
            to determine if this assessment is complete. There are currently
            no requirements for a self-assessment.
    Returns:
        True if the submitter has assessed their answer
    Examples:
        >>> submitter_is_finished('222bdf3d-a88e-11e3-859e-040ccee02800', {})
        True
    """
    return Assessment.objects.filter(
        score_type=SELF_TYPE, submission_uuid=submission_uuid
    ).exists()


def assessment_is_finished(submission_uuid, self_requirements):
    """
    Check whether a self-assessment has been completed. For self-assessment,
    this function is synonymous with submitter_is_finished.

    Args:
        submission_uuid (str): The unique identifier of the submission.
        self_requirements (dict): Any attributes of the assessment module required
            to determine if this assessment is complete. There are currently
            no requirements for a self-assessment.
    Returns:
        True if the assessment is complete.
    Examples:
        >>> assessment_is_finished('222bdf3d-a88e-11e3-859e-040ccee02800', {})
        True
    """
    return submitter_is_finished(submission_uuid, self_requirements)


def get_score(submission_uuid, self_requirements):  # pylint: disable=unused-argument
    """
    Get the score for this particular assessment.

    Args:
        submission_uuid (str): The unique identifier for the submission
        self_requirements (dict): Not used.
    Returns:
        A dictionary with the points earned, points possible, and
        contributing_assessments information, along with a None staff_id.
    Examples:
        >>> get_score('222bdf3d-a88e-11e3-859e-040ccee02800', {})
        {
            'points_earned': 5,
            'points_possible': 10
        }
    """
    assessment = get_assessment(submission_uuid)
    if not assessment:
        return None

    return {
        "points_earned": assessment["points_earned"],
        "points_possible": assessment["points_possible"],
        "contributing_assessments": [assessment["id"]],
        "staff_id": None,
    }


def create_assessment(
        submission_uuid,
        user_id,
        options_selected,
        criterion_feedback,
        overall_feedback,
        rubric_dict,
        scored_at=None
):
    """
    Create a self-assessment for a submission.

    Args:
        submission_uuid (str): The unique identifier for the submission being assessed.
        user_id (str): The ID of the user creating the assessment.
                       This must match the ID of the user who made the submission.
        options_selected (dict): Mapping of rubric criterion names to option values selected.
        criterion_feedback (dict): Dictionary mapping criterion names to the
            free-form text feedback the user gave for the criterion.
            Since criterion feedback is optional, some criteria may not appear
            in the dictionary.
        overall_feedback (unicode): Free-form text feedback on the submission overall.
        rubric_dict (dict): Serialized Rubric model.

    Keyword Arguments:
        scored_at (datetime): The timestamp of the assessment; defaults to the current time.

    Returns:
        dict: serialized Assessment model

    Raises:
        SelfAssessmentRequestError: Could not retrieve a submission that the user is allowed to score.
    """
    # Check that there are not any assessments for this submission
    if Assessment.objects.filter(submission_uuid=submission_uuid, score_type=SELF_TYPE).exists():
        msg = (
            "Cannot submit a self-assessment for the submission {uuid} "
            "because another self-assessment already exists for that submission."
        ).format(uuid=submission_uuid)
        raise SelfAssessmentRequestError(msg)

    # Check that the student is allowed to assess this submission
    try:
        submission = get_submission_and_student(submission_uuid)
        if submission['student_item']['student_id'] != user_id:
            msg = (
                "Cannot submit a self-assessment for the submission {uuid} "
                "because it was created by another learner "
                "(submission learner ID {student_id} does not match your "
                "learner id {other_id})"
            ).format(
                uuid=submission_uuid,
                student_id=submission['student_item']['student_id'],
                other_id=user_id
            )
            raise SelfAssessmentRequestError(msg)
    except SubmissionNotFoundError as ex:
        msg = (
            "Could not submit a self-assessment because no submission exists with UUID {uuid}"
        ).format(uuid=submission_uuid)
        raise SelfAssessmentRequestError() from ex

    try:
        assessment = _complete_assessment(
            submission_uuid,
            user_id,
            options_selected,
            criterion_feedback,
            overall_feedback,
            rubric_dict,
            scored_at
        )
        _log_assessment(assessment, submission)
    except InvalidRubric as ex:
        msg = "Invalid rubric definition: " + str(ex)
        logger.warning(msg, exc_info=True)
        raise SelfAssessmentRequestError(msg) from ex
    except InvalidRubricSelection as ex:
        msg = "Selected options do not match the rubric: " + str(ex)
        logger.warning(msg, exc_info=True)
        raise SelfAssessmentRequestError(msg) from ex
    except DatabaseError as ex:
        error_message = (
            "Error creating self assessment for submission {}"
        ).format(submission_uuid)
        logger.exception(error_message)
        raise SelfAssessmentInternalError(error_message) from ex

    # Return the serialized assessment
    return full_assessment_dict(assessment)


@transaction.atomic
def _complete_assessment(
        submission_uuid,
        user_id,
        options_selected,
        criterion_feedback,
        overall_feedback,
        rubric_dict,
        scored_at
):
    """
    Internal function for creating an assessment and its parts atomically.

    Args:
        submission_uuid (str): The unique identifier for the submission being
            assessed.
        user_id (str): The ID of the user creating the assessment. This must
            match the ID of the user who made the submission.
        options_selected (dict): Mapping of rubric criterion names to option
            values selected.
        criterion_feedback (dict): Dictionary mapping criterion names to the
            free-form text feedback the user gave for the criterion.
            Since criterion feedback is optional, some criteria may not appear
            in the dictionary.
        overall_feedback (unicode): Free-form text feedback on the submission overall.
        rubric_dict (dict): Serialized Rubric model.
        scored_at (datetime): The timestamp of the assessment.

    Returns:
        Assessment model

    """
    # Get or create the rubric
    rubric = rubric_from_dict(rubric_dict)

    # Create the self assessment
    assessment = Assessment.create(
        rubric,
        user_id,
        submission_uuid,
        SELF_TYPE,
        scored_at=scored_at,
        feedback=overall_feedback
    )

    # This will raise an `InvalidRubricSelection` if the selected options do not match the rubric.
    AssessmentPart.create_from_option_names(assessment, options_selected, feedback=criterion_feedback)
    return assessment


def get_assessment(submission_uuid):
    """
    Retrieve a self-assessment for a submission_uuid.

    Args:
        submission_uuid (str): The submission UUID for we want information for
            regarding self assessment.

    Returns:
        assessment (dict) is a serialized Assessment model, or None (if the user has not yet self-assessed)
        If multiple submissions or self-assessments are found, returns the most recent one.

    Raises:
        SelfAssessmentRequestError: submission_uuid was invalid.
    """
    # Retrieve assessments for the submission UUID
    # We weakly enforce that number of self-assessments per submission is <= 1,
    # but not at the database level.  Someone could take advantage of the race condition
    # between checking the number of self-assessments and creating a new self-assessment.
    # To be safe, we retrieve just the most recent submission.
    serialized_assessments = serialize_assessments(Assessment.objects.filter(
        score_type=SELF_TYPE, submission_uuid=submission_uuid
    ).order_by('-scored_at')[:1])

    if not serialized_assessments:
        logger.info("No self-assessment found for submission %s", submission_uuid)
        return None

    serialized_assessment = serialized_assessments[0]
    logger.info("Retrieved self-assessment for submission %s", submission_uuid)

    return serialized_assessment


def get_assessment_scores_by_criteria(submission_uuid):
    """Get the median score for each rubric criterion

    Args:
        submission_uuid (str): The submission uuid is used to get the
            assessments used to score this submission, and generate the
            appropriate median score.

    Returns:
        (dict): A dictionary of rubric criterion names, with a median score of
            the peer assessments.

    Raises:
        SelfAssessmentInternalError: If any error occurs while retrieving
            information to form the median scores, an error is raised.
    """
    try:
        # This will always create a list of length 1
        assessments = list(
            Assessment.objects.filter(
                score_type=SELF_TYPE, submission_uuid=submission_uuid
            ).order_by('-scored_at')[:1]
        )
        scores = Assessment.scores_by_criterion(assessments)
        # Since this is only being sent one score, the median score will be the
        # same as the only score.
        return Assessment.get_median_score_dict(scores)
    except DatabaseError as ex:
        error_message = (
            "Error getting self assessment scores for submission {}"
        ).format(submission_uuid)
        logger.exception(error_message)
        raise SelfAssessmentInternalError(error_message) from ex


def _log_assessment(assessment, submission):
    """
    Log the creation of a self-assessment.

    Args:
        assessment (Assessment): The assessment model.
        submission (dict): The serialized submission model.

    Returns:
        None

    """
    logger.info(
        "Created self-assessment %s for learner %s on submission %s, course %s, item %s with rubric %s",
        assessment.id,
        submission['student_item']['student_id'],
        submission['uuid'],
        submission['student_item']['course_id'],
        submission['student_item']['item_id'],
        assessment.rubric.content_hash
    )



"""
Studio editing view for OpenAssessment XBlock.
"""


import copy
import logging
from uuid import uuid4

from voluptuous import MultipleInvalid
from web_fragments.fragment import Fragment
from xblock.core import XBlock
from xblock.fields import List, Scope

from django.template.loader import get_template
from django.utils.translation import gettext_lazy

from openassessment.xblock.data_conversion import (create_rubric_dict, make_django_template_key,
                                                   update_assessments_format)
from openassessment.xblock.defaults import DEFAULT_EDITOR_ASSESSMENTS_ORDER, DEFAULT_RUBRIC_FEEDBACK_TEXT
from openassessment.xblock.editor_config import AVAILABLE_EDITORS
from openassessment.xblock.load_static import LoadStatic
from openassessment.xblock.resolve_dates import DateValidationError, InvalidDateFormat, parse_date_value, resolve_dates
from openassessment.xblock.schema import EDITOR_UPDATE_SCHEMA
from openassessment.xblock.validation import validator

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class StudioMixin:
    """
    Studio editing view for OpenAssessment XBlock.
    """

    DEFAULT_CRITERIA = [
        {
            'label': '',
            'options': [
                {
                    'label': ''
                },
            ]
        }
    ]

    NECESSITY_OPTIONS = {
        "required": gettext_lazy("Required"),
        "optional": gettext_lazy("Optional"),
        "": gettext_lazy("None")
    }

    # Build editor options from AVAILABLE_EDITORS
    AVAILABLE_EDITOR_OPTIONS = {
        key: val.get('display_name', key) for key, val in AVAILABLE_EDITORS.items()
    }

    STUDIO_EDITING_TEMPLATE = 'openassessmentblock/edit/oa_edit.html'

    BASE_EDITOR_ASSESSMENTS_ORDER = copy.deepcopy(DEFAULT_EDITOR_ASSESSMENTS_ORDER)

    # Since the XBlock problem definition contains only assessment
    # modules that are enabled, we need to keep track of the order
    # that the user left assessments in the editor, including
    # the ones that were disabled.  This allows us to keep the order
    # that the user specified.
    editor_assessments_order = List(
        default=DEFAULT_EDITOR_ASSESSMENTS_ORDER,
        scope=Scope.content,
        help="The order to display assessments in the editor."
    )

    def studio_view(self, context=None):  # pylint: disable=unused-argument
        """
        Render the OpenAssessment XBlock for editing in Studio.

        Args:
            context: Not actively used for this view.

        Returns:
            (Fragment): An HTML fragment for editing the configuration of this XBlock.
        """
        rendered_template = get_template(
            self.STUDIO_EDITING_TEMPLATE
        ).render(self.editor_context())
        fragment = Fragment(rendered_template)

        fragment.add_javascript_url(LoadStatic.get_url('openassessment-studio.js'))

        js_context_dict = {
            "ALLOWED_IMAGE_EXTENSIONS": self.ALLOWED_IMAGE_EXTENSIONS,
            "ALLOWED_FILE_EXTENSIONS": self.ALLOWED_FILE_EXTENSIONS,
            "FILE_EXT_BLACK_LIST": self.FILE_EXT_BLACK_LIST,
        }
        fragment.initialize_js('OpenAssessmentEditor', js_context_dict)
        return fragment

    def editor_context(self):
        """
        Update the XBlock's XML.

        Returns:
            dict with keys
                'rubric' (unicode), 'prompt' (unicode), 'title' (unicode),
                'submission_start' (unicode),  'submission_due' (unicode),
                'assessments (dict)

        """
        # In the authoring GUI, date and time fields should never be null.
        # Therefore, we need to resolve all "default" dates to datetime objects
        # before displaying them in the editor.
        try:
            __, __, date_ranges = resolve_dates(  # pylint: disable=redeclared-assigned-name
                self.start, self.due,
                [
                    (self.submission_start, self.submission_due)
                ] + [
                    (asmnt.get('start'), asmnt.get('due'))
                    for asmnt in self.valid_assessments
                ],
                self._
            )
        except (DateValidationError, InvalidDateFormat):
            # If the dates are somehow invalid, we still want users to be able to edit the ORA,
            # so just present the dates as they are.
            def _parse_date_safe(date):
                try:
                    return parse_date_value(date, self._)
                except InvalidDateFormat:
                    return ''

            date_ranges = [
                (_parse_date_safe(self.submission_start), _parse_date_safe(self.submission_due))
            ] + [
                (_parse_date_safe(asmnt.get('start')), _parse_date_safe(asmnt.get('due')))
                for asmnt in self.valid_assessments
            ]

        submission_start, submission_due = date_ranges[0]
        assessments = self._assessments_editor_context(date_ranges[1:])
        self.editor_assessments_order = self._editor_assessments_order_context()

        # Every rubric requires one criterion. If there is no criteria
        # configured for the XBlock, return one empty default criterion, with
        # an empty default option.
        criteria = copy.deepcopy(self.rubric_criteria_with_labels)
        if not criteria:
            criteria = self.DEFAULT_CRITERIA

        # To maintain backwards compatibility, if there is no
        # feedback_default_text configured for the xblock, use the default text
        feedback_default_text = copy.deepcopy(self.rubric_feedback_default_text)
        if not feedback_default_text:
            feedback_default_text = DEFAULT_RUBRIC_FEEDBACK_TEXT
        course_id = self.location.course_key if hasattr(self, 'location') else None

        # If allowed file types haven't been explicitly set, load from a preset
        white_listed_file_types = self.get_allowed_file_types_or_preset()
        white_listed_file_types_string = ','.join(white_listed_file_types) if white_listed_file_types else ''

        # If rubric reuse is enabled, include information about the other ORAs in this course
        rubric_reuse_data = {}
        if self.is_rubric_reuse_enabled:
            rubric_reuse_data = self.get_other_ora_blocks_for_rubric_editor_context()

        return {
            'prompts': self.prompts,
            'prompts_type': self.prompts_type,
            'title': self.title,
            'submission_due': submission_due,
            'submission_start': submission_start,
            'assessments': assessments,
            'criteria': criteria,
            'feedbackprompt': self.rubric_feedback_prompt,
            'feedback_default_text': feedback_default_text,
            'text_response': self.text_response if self.text_response else '',
            'text_response_editor': self.text_response_editor if self.text_response_editor else 'text',
            'file_upload_response': self.file_upload_response if self.file_upload_response else '',
            'necessity_options': self.NECESSITY_OPTIONS,
            'available_editor_options': self.AVAILABLE_EDITOR_OPTIONS,
            'file_upload_type': self.file_upload_type,
            'allow_multiple_files': self.allow_multiple_files,
            'white_listed_file_types': white_listed_file_types_string,
            'allow_latex': self.allow_latex,
            'leaderboard_show': self.leaderboard_show,
            'editor_assessments_order': [
                make_django_template_key(asmnt)
                for asmnt in self.editor_assessments_order
            ],
            'teams_feature_enabled': self.team_submissions_enabled,
            'teams_enabled': self.teams_enabled,
            'base_asset_url': self._get_base_url_path_for_course_assets(course_id),
            'is_released': self.is_released(),
            'teamsets': self.get_teamsets(course_id),
            'selected_teamset_id': self.selected_teamset_id,
            'show_rubric_during_response': self.show_rubric_during_response,
            'rubric_reuse_enabled': self.is_rubric_reuse_enabled,
            'rubric_reuse_data': rubric_reuse_data,
            'block_location': str(self.location),
        }

    @XBlock.json_handler
    def update_editor_context(self, data, suffix=''):  # pylint: disable=unused-argument
        """
        Update the XBlock's configuration.

        Args:
            data (dict): Data from the request; should have the format described
            in the editor schema.

        Keyword Arguments:
            suffix (str): Not used

        Returns:
            dict with keys 'success' (bool) and 'msg' (str)
        """

        # Validate and sanitize the data using a schema
        # If the data is invalid, this means something is wrong with
        # our JavaScript, so we log an exception.
        try:
            data = EDITOR_UPDATE_SCHEMA(data)
        except MultipleInvalid:
            logger.exception('Editor context is invalid')
            return {'success': False, 'msg': self._('Error updating XBlock configuration')}

        # Check that the editor assessment order contains all the assessments.
        current_order = set(data['editor_assessments_order'])
        if set(DEFAULT_EDITOR_ASSESSMENTS_ORDER) != current_order:
            # Backwards compatibility: "staff-assessment" may not be present.
            # If that is the only problem with this data, just add it manually and continue.
            if set(DEFAULT_EDITOR_ASSESSMENTS_ORDER) == current_order | {'staff-assessment'}:
                data['editor_assessments_order'].append('staff-assessment')
                logger.info('Backwards compatibility: editor_assessments_order now contains staff-assessment')
            else:
                logger.exception('editor_assessments_order does not contain all expected assessment types')
                return {'success': False, 'msg': self._('Error updating XBlock configuration')}

        if not data['text_response'] and not data['file_upload_response']:
            return {
                'success': False,
                'msg': self._("Error: Text Response and File Upload Response cannot both be disabled")
            }
        if not data['text_response'] and data['file_upload_response'] == 'optional':
            return {'success': False,
                    'msg': self._("Error: When Text Response is disabled, File Upload Response must be Required")}
        if not data['file_upload_response'] and data['text_response'] == 'optional':
            return {'success': False,
                    'msg': self._("Error: When File Upload Response is disabled, Text Response must be Required")}

        # Backwards compatibility: We used to treat "name" as both a user-facing label
        # and a unique identifier for criteria and options.
        # Now we treat "name" as a unique identifier, and we've added an additional "label"
        # field that we display to the user.
        # If the JavaScript editor sends us a criterion or option without a "name"
        # field, we should assign it a unique identifier.
        for criterion in data['criteria']:
            if 'name' not in criterion:
                criterion['name'] = uuid4().hex
            for option in criterion['options']:
                if 'name' not in option:
                    option['name'] = uuid4().hex

        xblock_validator = validator(self, self._)
        success, msg = xblock_validator(
            create_rubric_dict(data['prompts'], data['criteria']),
            data['assessments'],
            submission_start=data['submission_start'],
            submission_due=data['submission_due'],
            leaderboard_show=data['leaderboard_show']
        )
        if not success:
            return {'success': False, 'msg': self._('Validation error: {error}').format(error=msg)}

        # At this point, all the input data has been validated,
        # so we can safely modify the XBlock fields.
        self.title = data['title']
        self.display_name = data['title']
        self.prompts = data['prompts']
        self.prompts_type = data['prompts_type']
        self.rubric_criteria = data['criteria']
        self.rubric_assessments = data['assessments']
        self.editor_assessments_order = data['editor_assessments_order']
        self.rubric_feedback_prompt = data['feedback_prompt']
        self.rubric_feedback_default_text = data['feedback_default_text']
        self.submission_start = data['submission_start']
        self.submission_due = data['submission_due']
        self.text_response = data['text_response']
        self.text_response_editor = data['text_response_editor']
        self.file_upload_response = data['file_upload_response']
        if data['file_upload_response']:
            self.file_upload_type = data['file_upload_type']
            self.white_listed_file_types_string = data['white_listed_file_types']
        else:
            self.file_upload_type = None
            self.white_listed_file_types_string = None
        self.allow_multiple_files = bool(data['allow_multiple_files'])
        self.allow_latex = bool(data['allow_latex'])
        self.leaderboard_show = data['leaderboard_show']
        self.teams_enabled = bool(data.get('teams_enabled', False))
        self.selected_teamset_id = data.get('selected_teamset_id', '')
        self.show_rubric_during_response = data.get('show_rubric_during_response', False)

        return {'success': True, 'msg': self._('Successfully updated OpenAssessment XBlock')}

    @XBlock.json_handler
    def check_released(self, data, suffix=''):  # pylint: disable=unused-argument
        """
        Check whether the problem has been released.

        Args:
            data (dict): Not used

        Keyword Arguments:
            suffix (str): Not used

        Returns:
            dict with keys 'success' (bool), 'message' (unicode), and 'is_released' (bool)
        """
        # There aren't currently any server-side error conditions we report to the client,
        # but we send success/msg values anyway for consistency with other handlers.
        return {
            'success': True, 'msg': '',
            'is_released': self.is_released()
        }

    def _assessments_editor_context(self, assessment_dates):
        """
        Transform the rubric assessments list into the context
        we will pass to the Django template.

        Args:
            assessment_dates: List of assessment date ranges (tuples of start/end datetimes).

        Returns:
            dict

        """
        assessments = {}
        for asmnt, date_range in zip(self.rubric_assessments, assessment_dates):
            # Django Templates cannot handle dict keys with dashes, so we'll convert
            # the dashes to underscores.
            template_name = make_django_template_key(asmnt['name'])
            assessments[template_name] = copy.deepcopy(asmnt)
            assessments[template_name]['start'] = date_range[0]
            assessments[template_name]['due'] = date_range[1]

        # In addition to the data in the student training assessment, we need to include two additional
        # pieces of information: a blank context to render the empty template with, and the criteria
        # for each example (so we don't have any complicated logic within the template). Though this
        # could be accomplished within the template, we are opting to remove logic from the template.
        student_training_module = self.get_assessment_module('student-training')

        student_training_template = {
            'answer': {
                'parts': [
                    {'text': ''} for _ in self.prompts
                ]
            }
        }
        criteria_list = copy.deepcopy(self.rubric_criteria_with_labels)
        for criterion in criteria_list:
            criterion['option_selected'] = ""
        student_training_template['criteria'] = criteria_list

        if student_training_module:
            student_training_module = update_assessments_format([student_training_module])[0]
            example_list = []
            # Adds each example to a modified version of the student training module dictionary.
            for example in student_training_module['examples']:
                criteria_list = copy.deepcopy(self.rubric_criteria_with_labels)
                # Equivalent to a Join Query, this adds the selected option to the Criterion's dictionary, so that
                # it can be easily referenced in the template without searching through the selected options.
                for criterion in criteria_list:
                    for option_selected in example['options_selected']:
                        if option_selected['criterion'] == criterion['name']:
                            criterion['option_selected'] = option_selected['option']
                example_list.append({
                    'answer': example['answer'],
                    'criteria': criteria_list,
                })
            assessments['training'] = {'examples': example_list, 'template': student_training_template}
        # If we don't have student training enabled, we still need to render a single (empty, or default) example
        else:
            assessments['training'] = {'examples': [student_training_template], 'template': student_training_template}

        return assessments

    def _editor_assessments_order_context(self):
        """
        Create a list of assessment names in the order
        the user last set in the editor, including
        assessments that are not currently enabled.

        Returns:
            list of assessment names

        """
        # Start with the default order, to pick up any assessment types that have been added
        # since the user last saved their ordering.
        effective_order = copy.deepcopy(self.BASE_EDITOR_ASSESSMENTS_ORDER)

        # Account for changes the user has made to the default order
        user_order = copy.deepcopy(self.editor_assessments_order)
        effective_order = self._subset_in_relative_order(effective_order, user_order)

        # Account for inconsistencies between the user's order and the problems
        # that are currently enabled in the problem (These cannot be changed)
        enabled_assessments = [asmnt['name'] for asmnt in self.valid_assessments]
        enabled_ordered_assessments = [
            assessment for assessment in enabled_assessments if assessment in user_order
        ]
        effective_order = self._subset_in_relative_order(effective_order, enabled_ordered_assessments)

        return effective_order

    def _subset_in_relative_order(self, superset, subset):
        """
        Returns a copy of superset, with entries that appear in subset being reordered to match
        their relative ordering in subset.
        """
        superset_indices = [superset.index(item) for item in subset]
        sorted_superset_indices = sorted(superset_indices)
        if superset_indices != sorted_superset_indices:
            for index, superset_index in enumerate(sorted_superset_indices):
                superset[superset_index] = subset[index]
        return superset

    def _get_base_url_path_for_course_assets(self, course_key):
        """
        Returns base url path for course assets
        """
        if course_key is None:
            return None

        placeholder_id = uuid4().hex
        # create a dummy asset location with a fake but unique name. strip off the name, and return it
        url_path = str(course_key.make_asset_key('asset', placeholder_id).for_branch(None))
        if not url_path.startswith('/'):
            url_path = '/' + url_path
        return url_path.replace(placeholder_id, '')

    def get_team_configuration(self, course_id):
        """
        Returns a dict with team configuration settings.
        """
        configuration_service = self.runtime.service(self, 'teams_configuration')
        team_configuration = configuration_service.get_teams_configuration(course_id)
        if not team_configuration:
            return None
        return team_configuration

    def get_teamsets(self, course_id):
        """
        Wrapper around get_team_configuration that returns team names only for display
        """
        team_configuration = self.get_team_configuration(course_id)
        if not team_configuration:
            return None
        return team_configuration.teamsets



"""
Django models for training (both student and AI).
"""


from hashlib import sha1
import json

from django.core.cache import cache
from django.db import models

from .base import CriterionOption, Rubric


class TrainingExample(models.Model):
    """
    An example assessment used to train students (before peer assessment) or AI.
    """
    # The answer (JSON-serialized)
    raw_answer = models.TextField(blank=True)

    rubric = models.ForeignKey(Rubric, on_delete=models.CASCADE)

    # Use a m2m to avoid changing the criterion option
    options_selected = models.ManyToManyField(CriterionOption)

    # SHA1 hash
    content_hash = models.CharField(max_length=40, unique=True, db_index=True)

    class Meta:
        app_label = "assessment"

    @classmethod
    def create_example(cls, answer, options_selected, rubric):
        """
        Create a new training example.

        Args:
            answer (JSON-serializable): The answer associated with the training example.
            options_selected (dict): The options selected from the rubric (mapping of criterion names to option names)
            rubric (Rubric): The rubric associated with the training example.

        Returns:
            TrainingExample

        Raises:
            InvalidRubricSelection

        """
        content_hash = cls.calculate_hash(answer, options_selected, rubric)
        example = TrainingExample.objects.create(
            content_hash=content_hash,
            raw_answer=json.dumps(answer),
            rubric=rubric
        )

        # This will raise `InvalidRubricSelection` if the selected options
        # do not match the rubric.
        for criterion_name, option_name in options_selected.items():
            option = rubric.index.find_option(criterion_name, option_name)
            example.options_selected.add(option)
        return example

    @property
    def answer(self):
        """
        Return the JSON-decoded answer.

        Returns:
            JSON-serializable

        """
        return json.loads(self.raw_answer)

    @property
    def options_selected_dict(self):
        """
        Return a dictionary of the rubric options selected.

        Returns:
            dict: maps criterion names to selected option names

        """
        # Since training examples are immutable, we can safely cache this
        cache_key = self.cache_key_serialized(attribute="options_selected_dict")
        options_selected = cache.get(cache_key)
        if options_selected is None:
            options_selected = {
                option.criterion.name: option.name
                for option in self.options_selected.all()
            }
            cache.set(cache_key, options_selected)
        return options_selected

    def cache_key_serialized(self, attribute=None):
        """
        Create a cache key based on the content hash
        for serialized versions of this model.

        Keyword Arguments:
            attribute: The name of the attribute being serialized.
                If not specified, assume that we are serializing the entire model.

        Returns:
            str: The cache key

        """
        if attribute is None:
            key_template = "TrainingExample.json.{content_hash}"
        else:
            key_template = "TrainingExample.{attribute}.json.{content_hash}"

        cache_key = key_template.format(
            content_hash=self.content_hash,
            attribute=attribute
        )
        return cache_key

    @staticmethod
    def calculate_hash(answer, options_selected, rubric):
        """
        Calculate a hash for the contents of training example.

        Args:
            answer (JSON-serializable): The answer associated with the training example.
            options_selected (dict): The options selected from the rubric (mapping of criterion names to option names)
            rubric (Rubric): The rubric associated with the training example.

        Returns:
            str

        """
        contents = json.dumps({
            'answer': answer,
            'options_selected': options_selected,
            'rubric': rubric.id
        })
        return sha1(contents.encode('utf-8')).hexdigest()

    @classmethod
    def cache_key(cls, answer, options_selected, rubric):
        """
        Calculate a cache key based on the content hash.

        Args:
            answer (JSON-serializable): The answer associated with the training example.
            options_selected (dict): The options selected from the rubric (mapping of criterion names to option names)
            rubric (Rubric): The rubric associated with the training example.

        Returns:
            tuple of `(cache_key, content_hash)`, both bytestrings

        """
        content_hash = cls.calculate_hash(answer, options_selected, rubric)
        cache_key = "TrainingExample.model.{content_hash}".format(
            content_hash=content_hash
        )
        return cache_key, content_hash



"""
API endpoints for enhanced staff grader
"""
from functools import wraps
import logging

from django.db.models import Case, OuterRef, Prefetch, Subquery, Value, When
from django.db.models.fields import CharField
from submissions.api import get_student_ids_by_submission_uuid, get_submission
from submissions.errors import SubmissionInternalError, SubmissionNotFoundError, SubmissionRequestError, SubmissionError
from submissions.team_api import get_team_ids_by_team_submission_uuid, get_team_submission
from xblock.core import XBlock
from xblock.exceptions import JsonHandlerError

from openassessment.assessment.models.base import Assessment, AssessmentPart
from openassessment.assessment.models.staff import StaffWorkflow, TeamStaffWorkflow
from openassessment.data import map_anonymized_ids_to_usernames, OraSubmissionAnswerFactory, VersionNotFoundException
from openassessment.staffgrader.errors.submission_lock import SubmissionLockContestedError
from openassessment.staffgrader.models.submission_lock import SubmissionGradingLock
from openassessment.staffgrader.serializers import (
    AssessmentSerializer,
    MissingContextException,
    SubmissionDetailFileSerilaizer,
    SubmissionListSerializer,
    SubmissionLockSerializer,
    TeamSubmissionListSerializer,
)
from openassessment.xblock.staff_area_mixin import require_course_staff


log = logging.getLogger(__name__)


def require_submission_uuid(validate=True):
    """
    Unpacks and passes submission_uuid from request to handler function.

    params:
    - validate: Whether or not to check submissions to see if this is a real submission UUID or not. Default True

    Raises:
    - 400 if the submission_uuid was not provided or was incorrectly formatted
    - 404 if the submission_uuid wasn't found in submissions
    - 500 for errors with submissions or general exceptions
    """
    def decorator(handler):
        @wraps(handler)
        def wrapped_handler(self, data, suffix=""):  # pylint: disable=unused-argument
            submission_uuid = data.get('submission_uuid', None)
            if not submission_uuid:
                raise JsonHandlerError(400, "Body must contain a submission_uuid")
            if validate:
                try:
                    if self.is_team_assignment():
                        get_team_submission(submission_uuid)
                    else:
                        get_submission(submission_uuid)
                except SubmissionNotFoundError as exc:
                    raise JsonHandlerError(404, "Submission not found") from exc
                except SubmissionRequestError as exc:
                    raise JsonHandlerError(400, "Bad submission_uuid provided") from exc
                except (SubmissionInternalError, Exception) as exc:
                    raise JsonHandlerError(500, "Internal error getting submission info") from exc
            return handler(self, submission_uuid, data, suffix=suffix)
        return wrapped_handler
    return decorator


class StaffGraderMixin:
    """
    Actions to interact with submission locks, blocking other staff from grading assignments while
    grading is in progress.
    """

    @XBlock.json_handler
    @require_course_staff("STUDENT_GRADE")
    @require_submission_uuid(validate=False)
    def check_submission_lock(self, submission_uuid, data, suffix=""):  # pylint: disable=unused-argument
        """
        Get info about a submission lock. Does not verify that the ID is a valid submission.

        Returns:
        - Serialized submission lock info.
        """
        anonymous_user_id = self.get_anonymous_user_id_from_xmodule_runtime()
        context = {'user_id': anonymous_user_id}

        submission_lock = SubmissionGradingLock.get_submission_lock(submission_uuid) or {}
        return SubmissionLockSerializer(submission_lock, context=context).data

    @XBlock.json_handler
    @require_course_staff("STUDENT_GRADE")
    @require_submission_uuid(validate=True)
    def claim_submission_lock(self, submission_uuid, data, suffix=''):  # pylint: disable=unused-argument
        """
        Attempt to claim or reclaim a submission lock

        Returns:
        - Serialized submission lock info.

        Raises:
        - 403 in the case of a contested lock
        """
        anonymous_user_id = self.get_anonymous_user_id_from_xmodule_runtime()
        context = {'user_id': anonymous_user_id}

        try:
            submission_lock = SubmissionGradingLock.claim_submission_lock(submission_uuid, anonymous_user_id)
            return SubmissionLockSerializer(submission_lock, context=context).data
        except SubmissionLockContestedError as err:
            raise JsonHandlerError(403, err.get_error_code()) from err

    @XBlock.json_handler
    @require_course_staff("STUDENT_GRADE")
    @require_submission_uuid(validate=False)
    def delete_submission_lock(self, submission_uuid, data, suffix=''):  # pylint: disable=unused-argument
        """
        Attempt to clear a submission lock.

        Returns:
        - Serialized submission lock info (which in this case would just be {'lock_status': 'unlocked'})

        Raises:
        - 403 in the case of a contested lock
        """
        anonymous_user_id = self.get_anonymous_user_id_from_xmodule_runtime()
        context = {'user_id': anonymous_user_id}

        try:
            SubmissionGradingLock.clear_submission_lock(submission_uuid, anonymous_user_id)
            return SubmissionLockSerializer({}, context=context).data
        except SubmissionLockContestedError as err:
            raise JsonHandlerError(403, err.get_error_code()) from err

    @XBlock.json_handler
    @require_course_staff("STUDENT_GRADE")
    def list_staff_workflows(self, data, suffix=''):  # pylint: disable=unused-argument
        """
        Returns data for the base "list" view, showing a summary of all graded / gradable items in the given assignment

        Example Data Shape:
        {
            submission_uuid:
        }
        """
        # Calculate this once so we don't have to re-query each time
        is_team_assignment = self.is_team_assignment()

        # Fetch staff workflows, annotated with grading_status and lock_status
        staff_workflows = self._bulk_fetch_annotated_staff_workflows(is_team_assignment=is_team_assignment)

        # Lookup additional info like usernames and assessments and determine serializer type
        serializer = TeamSubmissionListSerializer if is_team_assignment else SubmissionListSerializer
        serializer_context = self._get_list_workflows_serializer_context(
            staff_workflows, is_team_assignment=is_team_assignment
        )

        # Serialize workflows with the context, and return the dict of submissions
        result = {}
        for staff_workflow in staff_workflows:
            try:
                serialized_workflow = serializer(staff_workflow, context=serializer_context).data
                result[staff_workflow.identifying_uuid] = serialized_workflow
            except MissingContextException as e:
                log.exception("Failed to serialize workflow %d: %s", staff_workflow.id, str(e), exc_info=True)
        return result

    def _get_list_workflows_serializer_context(self, staff_workflows, is_team_assignment=False):
        """
        Fetch additional required data and models to serialize the response
        """
        # Pull out three sets from the workflows for use later
        submission_uuids, workflow_scorer_ids, assessment_ids = set(), set(), set()
        for workflow in staff_workflows:
            submission_uuids.add(workflow.identifying_uuid)
            if workflow.assessment:
                assessment_ids.add(workflow.assessment)
            if workflow.scorer_id:
                workflow_scorer_ids.add(workflow.scorer_id)
        course_id = self.get_student_item_dict()['course_id']

        # Fetch user identifier mappings
        if is_team_assignment:
            # Look up the team IDs for submissions so we can later map to team names
            team_submission_uuid_to_team_id = get_team_ids_by_team_submission_uuid(submission_uuids)

            # Look up names for teams
            topic_id = self.selected_teamset_id
            team_id_to_team_name = self.teams_service.get_team_names(course_id, topic_id)

            # Do bulk lookup for scorer anonymous ids (submitting team name is a separate lookup)
            anonymous_id_to_username = map_anonymized_ids_to_usernames(set(workflow_scorer_ids))

            context = {
                'team_submission_uuid_to_team_id': team_submission_uuid_to_team_id,
                'team_id_to_team_name': team_id_to_team_name,
            }
        else:
            # When we look up usernames we want to include all connected learner student ids
            submission_uuid_to_student_id = get_student_ids_by_submission_uuid(
                course_id,
                submission_uuids,
            )

            # Do bulk lookup for all anonymous ids (submitters and scoreres). This is used for the
            # `gradedBy` and `username` fields
            anonymous_id_to_username = map_anonymized_ids_to_usernames(
                set(submission_uuid_to_student_id.values()) | workflow_scorer_ids
            )

            context = {
                'submission_uuid_to_student_id': submission_uuid_to_student_id,
            }

        # Do a bulk fetch of the assessments linked to the workflows, including all connected
        # Rubric, Criteria, and Option models
        submission_uuid_to_assessment = self.bulk_deep_fetch_assessments(assessment_ids)

        context.update({
            'anonymous_id_to_username': anonymous_id_to_username,
            'submission_uuid_to_assessment': submission_uuid_to_assessment,
        })

        return context

    def _bulk_fetch_annotated_staff_workflows(self, is_team_assignment=False):
        """
        Returns: QuerySet of StaffWorkflows, filtered by the current course and item, with the following annotations:
         - current_lock_user: The "owner_id" of the most recent active (created less than TIME_LIMIT ago) lock
         - grading_status: one of
                              * "graded"   - the StaffWorkflow has an associated Assessment
                              * "ungraded" - the StaffWorkflow has no asociated Assessment
        - lock_status: one of
                              * "in-progress" - current_lock_user is the current user's anonymous id.
                                                The current user has an active lock on this submission.
                              * "locked"      - current_lock_user is non-null and not the current user's anonymous id.
                                                Another user has an active lock on this submission.
                              * "unlocked"    - current_lock_user is null
                                                There is no active lock on this submission.
        """
        # Create an unevaluated QuerySet of "active" SubmissionLock objects that refer to the same submission as the
        # "current" workflow
        student_item_dict = self.get_student_item_dict()

        # Return TeamStaffWorkflows for teams, StaffWorkflows for individual
        if is_team_assignment:
            workflow_type = TeamStaffWorkflow
            identifying_uuid = 'team_submission_uuid'
        else:
            workflow_type = StaffWorkflow
            identifying_uuid = 'submission_uuid'

        newest_lock = SubmissionGradingLock.currently_active().filter(
            submission_uuid=OuterRef(identifying_uuid)
        ).order_by(
            '-created_at'
        )

        staff_workflows = workflow_type.objects.filter(
            course_id=student_item_dict['course_id'],
            item_id=student_item_dict['item_id'],
            cancelled_at=None,
        ).annotate(
            current_lock_user=Subquery(newest_lock.values('owner_id')),
        ).annotate(
            grading_status=Case(
                When(assessment__isnull=False, then=Value("graded", output_field=CharField())),
                default=Value("ungraded", output_field=CharField())
            ),
            lock_status=Case(
                When(
                    current_lock_user=student_item_dict['student_id'],
                    then=Value("in-progress", output_field=CharField())
                ),
                When(
                    current_lock_user__isnull=False,
                    then=Value("locked", output_field=CharField())
                ),
                default=Value("unlocked", output_field=CharField())
            )
        )
        return staff_workflows

    def bulk_deep_fetch_assessments(self, assessment_ids):
        """
        Given a list of Assessment ids, fetch Assessments and prefetch
        linked Rubrics, AssessmentParts, Criteria, and Options.

        returns: (dict) mapping submission uuids to the associated assessment.
        If there is no assessment associated with a submission, it is not included in the dict.
        """
        assessments = Assessment.objects.filter(
            pk__in=assessment_ids
        ).prefetch_related(
            Prefetch(
                "parts",
                queryset=AssessmentPart.objects.select_related('criterion', 'option')
            ),
            "rubric__criteria",
            "rubric__criteria__options"

        ).select_related(
            'rubric',
        ).order_by('-scored_at')
        assessments_by_submission_uuid = {
            assessment.submission_uuid: assessment
            for assessment in assessments
        }
        return assessments_by_submission_uuid

    @XBlock.json_handler
    @require_course_staff("STUDENT_GRADE")
    @require_submission_uuid(validate=True)
    def get_submission_info(self, submission_uuid, _, suffix=''):  # pylint: disable=unused-argument
        """
        Return a dict representation of a submission in the form
        {
            'text': <list of strings representing the raw response for each prompt>
            'files': <list of:>
                {
                    'download_url': <file url>
                    'description': <file description>
                    'name': <file name>
                }
        }
        """
        try:
            if self.is_team_assignment():
                submission = get_team_submission(submission_uuid)
            else:
                submission = get_submission(submission_uuid)
            answer = OraSubmissionAnswerFactory.parse_submission_raw_answer(submission.get('answer'))
        except SubmissionError as err:
            raise JsonHandlerError(404, str(err)) from err
        except VersionNotFoundException as err:
            raise JsonHandlerError(500, str(err)) from err

        return {
            'files': [
                SubmissionDetailFileSerilaizer(file_data).data
                for file_data in self.get_download_urls_from_submission(submission)
            ],
            'text': answer.get_text_responses()
        }

    @XBlock.json_handler
    @require_course_staff("STUDENT_GRADE")
    @require_submission_uuid(validate=True)
    def get_assessment_info(self, submission_uuid, _, suffix=''):  # pylint: disable=unused-argument
        """
        Returns a dict representation of a staff assessment in the form
        {
            'feedback': <submission-level feedback>
            'points_earned': <earned points>
            'points_possible': <maximum possible points>
            'criteria': list of {
                'name': <criterion name>
                'option': <name of selected option> This may be blank.
                          If so, there are no options defined for the given criterion and it is feedback-only
                'feedback': <feedback for criterion>
            }
        }
        """
        student_item_dict = self.get_student_item_dict()
        course_id = student_item_dict['course_id']
        item_id = student_item_dict['item_id']

        try:
            if self.is_team_assignment():
                workflow = TeamStaffWorkflow.get_team_staff_workflow(course_id, item_id, submission_uuid)
            else:
                workflow = StaffWorkflow.get_staff_workflow(course_id, item_id, submission_uuid)
        except StaffWorkflow.DoesNotExist as ex:
            msg = f"No gradeable submission found with uuid={submission_uuid} in course={course_id} item={item_id}"
            raise JsonHandlerError(404, msg) from ex

        if not workflow.assessment:
            return {}

        assessments = self.bulk_deep_fetch_assessments([workflow.assessment])
        if len(assessments) != 1 or submission_uuid not in assessments:
            log.error(
                (
                    "[%s] Error looking up assessments. Submission UUID = %s, "
                    "Staff Workflow Id = %d, Staff Workflow Assessment = %s, Assessments = %s"
                ),
                item_id, submission_uuid, workflow.id, workflow.assessment, assessments
            )
            raise JsonHandlerError(500, "Error looking up assessments")

        assessment = assessments[submission_uuid]
        return AssessmentSerializer(assessment).data



"""
Handle OpenAssessment XBlock requests to the Workflow API.
"""

from xblock.core import XBlock
from submissions.api import get_submissions, SubmissionInternalError, SubmissionNotFoundError

from openassessment.workflow import api as workflow_api
from openassessment.workflow.models import AssessmentWorkflowCancellation


class WorkflowMixin:
    """
    Handle OpenAssessment XBlock requests to the Workflow API.
    """

    # Dictionary mapping assessment names (e.g. peer-assessment)
    # to the corresponding workflow step names.
    ASSESSMENT_STEP_NAMES = {
        "self-assessment": "self",
        "peer-assessment": "peer",
        "student-training": "training",
        "staff-assessment": "staff"
    }

    @XBlock.json_handler
    def handle_workflow_info(self, data, suffix=''):    # pylint:disable=W0613
        """
        Retrieve the current state of the workflow.

        Args:
            data: Unused

        Keyword Arguments:
            suffix: Unused

        Returns:
            dict

        """
        return self.get_workflow_info()

    def create_workflow(self, submission_uuid):
        """
        Create a new workflow for a student submission.

        Args:
            submission_uuid (str): The UUID of the submission to associate
                with the workflow.

        Returns:
            None

        """
        steps = self._create_step_list()
        workflow_api.create_workflow(submission_uuid, steps, on_init_params={})

    def workflow_requirements(self):
        """
        Retrieve the requirements from each assessment module
        so the workflow can decide whether the student can receive a score.

        Returns:
            dict

        """
        requirements = {}

        peer_assessment_module = self.get_assessment_module('peer-assessment')
        if peer_assessment_module:
            requirements["peer"] = {
                "must_grade": peer_assessment_module["must_grade"],
                "must_be_graded_by": peer_assessment_module["must_be_graded_by"],
                "enable_flexible_grading": peer_assessment_module.get("enable_flexible_grading", False)
            }

        training_module = self.get_assessment_module('student-training')
        if training_module:
            requirements["training"] = {
                "num_required": len(training_module["examples"])
            }

        staff_assessment_module = self.get_assessment_module('staff-assessment')
        if staff_assessment_module:
            requirements["staff"] = {
                "required": staff_assessment_module["required"]
            }

        return requirements

    def update_workflow_status(self, submission_uuid=None):
        """
        Update the status of a workflow.  For example, change the status
        from peer-assessment to self-assessment.  Creates a score
        if the student has completed all requirements.

        Keyword Arguments:
            submission_uuid (str): The submission associated with the workflow to update.
                Defaults to the submission created by the current student.

        Returns:
            None

        Raises:
            AssessmentWorkflowError
        """
        if submission_uuid is None:
            submission_uuid = self.submission_uuid

        if submission_uuid is not None:
            requirements = self.workflow_requirements()
            workflow_api.update_from_assessments(submission_uuid, requirements)

    def get_workflow_info(self, submission_uuid=None):
        """
        Retrieve a description of the student's progress in a workflow.
        Note that this *may* update the workflow status if it's changed.

        Keyword Arguments:
            submission_uuid (str): The submission associated with the workflow to return.
                Defaults to the submission created by the current student.

        Returns:
            dict

        Raises:
            AssessmentWorkflowError
        """
        if self.is_team_assignment():
            if submission_uuid:
                team_submission_uuid = self.get_team_submission_uuid_from_individual_submission_uuid(submission_uuid)
            else:
                team_submission_uuid = None
            return self.get_team_workflow_info(team_submission_uuid)

        if submission_uuid is None:
            submission_uuid = self.get_submission_uuid()

        if submission_uuid is None:
            return {}
        return workflow_api.get_workflow_for_submission(
            submission_uuid, self.workflow_requirements()
        )

    def get_submission_uuid(self):
        """ Submission UUIDs can be in multiple spots based on the submission type,
            try the various locations to try to find it.

            No submission ID will be found if a learner has not submitted a response

            Indiviual submissions will be in the user's context.

            Returns:
                (string) Submission ID if found
                (None) None if not found
        """
        if self.submission_uuid is not None:
            return self.submission_uuid
        elif self.is_team_assignment():
            try:
                # Query for submissions by the student item
                student_item = self.get_student_item_dict()
                submission_list = get_submissions(student_item)
                if submission_list and submission_list[0]["uuid"] is not None:
                    return submission_list[0]["uuid"]
            except (SubmissionInternalError, SubmissionNotFoundError):
                return None
        return None

    def get_workflow_status_counts(self):
        """
        Retrieve the counts of students in each step of the workflow.

        Returns:
            tuple of (list, int), where the list contains dicts with keys
            "status" (unicode value) and "count" (int value), and the
            integer represents the total number of submissions.

        Example Usage:
            >>> status_counts, num_submissions = xblock.get_workflow_status_counts()
            >>> num_submissions
                12
            >>> status_counts
                [
                    {"status": "peer", "count": 2},
                    {"status": "self", "count": 1},
                    {"status": "waiting": "count": 4},
                    {"status": "done", "count": 5}
                ]
        """
        student_item = self.get_student_item_dict()
        status_counts = workflow_api.get_status_counts(
            course_id=student_item['course_id'],
            item_id=student_item['item_id'],
            steps=self._create_step_list(),
        )
        num_submissions = sum(item['count'] for item in status_counts)
        return status_counts, num_submissions

    def _create_step_list(self):
        """
        Return a list of valid workflow step names.
        This translates between the assessment types (loaded from the problem definition)
        and the step types (used by the Workflow API).
        At some point, we should probably refactor to make these two names consistent.

        Returns:
            list

        """
        return [
            self.ASSESSMENT_STEP_NAMES.get(ra['name'])
            for ra in self.valid_assessments
            if ra['name'] in self.ASSESSMENT_STEP_NAMES
        ]

    def get_workflow_cancellation_info(self, submission_uuid):
        """
        Returns cancellation information for a particular submission.

        :param submission_uuid: The submission to return information for.
        :return: The cancellation information, or None if the submission has
        not been cancelled.
        """
        cancellation_info = workflow_api.get_assessment_workflow_cancellation(submission_uuid)
        if not cancellation_info:
            return None

        # Add the username of the staff member who cancelled the submission
        cancellation_info['cancelled_by'] = self.get_username(cancellation_info['cancelled_by_id'])

        # Add the date that the workflow was cancelled (in preference to the serialized date string)
        del cancellation_info['created_at']
        cancellation_model = AssessmentWorkflowCancellation.get_latest_workflow_cancellation(submission_uuid)
        if cancellation_model:
            cancellation_info['cancelled_at'] = cancellation_model.created_at

        return cancellation_info



"""Default data initializations for the XBlock, with formatting preserved."""
# pylint: disable=line-too-long

DEFAULT_PROMPT = """
    Censorship in the Libraries

    'All of us can think of a book that we hope none of our children or any other children have taken off the shelf. But if I have the right to remove that book from the shelf -- that work I abhor -- then you also have exactly the same right and so does everyone else. And then we have no books left on the shelf for any of us.' --Katherine Paterson, Author

    Write a persuasive essay to a newspaper reflecting your views on censorship in libraries. Do you believe that certain materials, such as books, music, movies, magazines, etc., should be removed from the shelves if they are found offensive? Support your position with convincing arguments from your own experience, observations, and/or reading.

    Read for conciseness, clarity of thought, and form.
"""  # nopep8

DEFAULT_RUBRIC_CRITERIA = [
    {
        'name': "Ideas",
        'label': "Ideas",
        'prompt': "Determine if there is a unifying theme or main idea.",
        'order_num': 0,
        'feedback': 'optional',
        'options': [
            {
                'order_num': 0, 'points': 0, 'name': 'Poor', 'label': 'Poor',
                'explanation': """Difficult for the reader to discern the main idea.  Too brief or too repetitive to establish or maintain a focus."""  # nopep8
            },
            {
                'order_num': 1, 'points': 3, 'name': 'Fair', 'label': 'Fair',
                'explanation': """Presents a unifying theme or main idea, but may include minor tangents.  Stays somewhat focused on topic and task."""  # nopep8
            },
            {
                'order_num': 2, 'points': 5, 'name': 'Good', 'label': 'Good',
                'explanation': """Presents a unifying theme or main idea without going off on tangents.  Stays completely focused on topic and task."""  # nopep8
            },
        ],
    },
    {
        'name': "Content",
        'label': "Content",
        'prompt': "Assess the content of the submission",
        'order_num': 1,
        'options': [
            {
                'order_num': 0, 'points': 0, 'name': 'Poor', 'label': 'Poor',
                'explanation': """Includes little information with few or no details or unrelated details.  Unsuccessful in attempts to explore any facets of the topic."""  # nopep8
            },
            {
                'order_num': 1, 'points': 1, 'name': 'Fair', 'label': 'Fair',
                'explanation': """Includes little information and few or no details.  Explores only one or two facets of the topic."""  # nopep8
            },
            {
                'order_num': 2, 'points': 3, 'name': 'Good', 'label': 'Good',
                'explanation': """Includes sufficient information and supporting details. (Details may not be fully developed; ideas may be listed.)  Explores some facets of the topic."""  # nopep8
            },
            {
                'order_num': 3, 'points': 5, 'name': 'Excellent', 'label': 'Excellent',
                'explanation': """Includes in-depth information and exceptional supporting details that are fully developed.  Explores all facets of the topic."""  # nopep8
            },
        ],
    },
]

# The rubric's feedback prompt is a set of instructions letting the student
# know they can provide additional free form feedback in their assessment.
DEFAULT_RUBRIC_FEEDBACK_PROMPT = """
(Optional) What aspects of this response stood out to you? What did it do well? How could it be improved?
"""

# The rubric's feedback text is the default text displayed and used as
# the student's response to the feedback prompt
DEFAULT_RUBRIC_FEEDBACK_TEXT = """
I think that this response...
"""

DEFAULT_EXAMPLE_ANSWER = (
    "Replace this text with your own sample response for this assignment. "
    "Then, under Response Score to the right, select an option for each criterion. "
    "Learners practice performing peer assessments by assessing this response and comparing "
    "the options that they select in the rubric with the options that you specified."
)

DEFAULT_EXAMPLE_ANSWER_2 = (
    "Replace this text with another sample response, "
    "and then specify the options that you would select for this response."
)

DEFAULT_STUDENT_TRAINING = {
    "name": "student-training",
    "start": None,
    "due": None,
    "examples": [
        {
            "answer": DEFAULT_EXAMPLE_ANSWER,
            "options_selected": [
                {
                    "criterion": "Ideas",
                    "option": "Fair"
                },
                {
                    "criterion": "Content",
                    "option": "Good"
                }
            ]
        },
        {
            "answer": DEFAULT_EXAMPLE_ANSWER_2,
            "options_selected": [
                {
                    "criterion": "Ideas",
                    "option": "Poor"
                },
                {
                    "criterion": "Content",
                    "option": "Good"
                }
            ]
        }
    ]
}

DEFAULT_START = "2001-01-01T00:00"
DEFAULT_DUE = "2029-01-01T00:00"

# The Default Peer Assessment is created as an example of how this XBlock can be
# configured. If no configuration is specified, this is the default assessment
# module(s) associated with the XBlock.
DEFAULT_PEER_ASSESSMENT = {
    "name": "peer-assessment",
    "start": DEFAULT_START,
    "due": DEFAULT_DUE,
    "must_grade": 5,
    "must_be_graded_by": 3,
    "enable_flexible_grading": False,
    "flexible_grading_days": 7,
    "flexible_grading_graded_by_percentage": 30
}

DEFAULT_SELF_ASSESSMENT = {
    "name": "self-assessment",
    "start": DEFAULT_START,
    "due": DEFAULT_DUE,
}

DEFAULT_STAFF_ASSESSMENT = {
    "name": "staff-assessment",
    "start": DEFAULT_START,
    "due": DEFAULT_DUE,
    "required": False,
}

ACTIVE_STAFF_ASSESSMENT = {
    "name": "staff-assessment",
    "start": DEFAULT_START,
    "due": DEFAULT_DUE,
    "required": True,
}

DEFAULT_ASSESSMENT_MODULES = [
    DEFAULT_STUDENT_TRAINING,
    DEFAULT_PEER_ASSESSMENT,
    DEFAULT_SELF_ASSESSMENT,
    DEFAULT_STAFF_ASSESSMENT,
]

DEFAULT_EDITOR_ASSESSMENTS_ORDER = [
    "student-training",
    "peer-assessment",
    "self-assessment",
    "staff-assessment",
]

BLANK_ASSESSMENT_MODULES = []

SELF_ASSESSMENT_MODULES = [
    DEFAULT_SELF_ASSESSMENT,
]

PEER_ASSESSMENT_MODULES = [
    DEFAULT_STUDENT_TRAINING,
    DEFAULT_PEER_ASSESSMENT,
]

STAFF_ASSESSMENT_MODULES = [
    ACTIVE_STAFF_ASSESSMENT,
]

SELF_TO_PEER_ASSESSMENT_MODULES = [
    DEFAULT_STUDENT_TRAINING,
    DEFAULT_SELF_ASSESSMENT,
    DEFAULT_PEER_ASSESSMENT,
]

SELF_TO_STAFF_ASSESSMENT_MODULES = [
    DEFAULT_SELF_ASSESSMENT,
    ACTIVE_STAFF_ASSESSMENT,
]



"""
Mixin for determining configuration and feature-toggle state relevant to an ORA block.
"""


from edx_django_utils.monitoring import set_custom_attribute
from edx_toggles.toggles import WaffleSwitch

from django.conf import settings
from django.utils.functional import cached_property

WAFFLE_NAMESPACE = 'openresponseassessment'

ALL_FILES_URLS = 'all_files_urls'
TEAM_SUBMISSIONS = 'team_submissions'
USER_STATE_UPLOAD_DATA = 'user_state_upload_data'
RUBRIC_REUSE = 'rubric_reuse'
ENHANCED_STAFF_GRADER = 'enhanced_staff_grader'

FEATURE_TOGGLES_BY_FLAG_NAME = {
    ALL_FILES_URLS: 'ENABLE_ORA_ALL_FILE_URLS',
    TEAM_SUBMISSIONS: 'ENABLE_ORA_TEAM_SUBMISSIONS',
    USER_STATE_UPLOAD_DATA: 'ENABLE_ORA_USER_STATE_UPLOAD_DATA',
    RUBRIC_REUSE: 'ENABLE_ORA_RUBRIC_REUSE',
    ENHANCED_STAFF_GRADER: 'ENABLE_ENHANCED_STAFF_GRADER'
}


def import_waffle_switch():
    """
    Helper method that imports WaffleSwitch from edx-platform at runtime.
    WARNING: This method is now deprecated and should not be relied upon.
    """
    set_custom_attribute("deprecated_edx_ora2", "import_waffle_switch")
    return WaffleSwitch


def import_course_waffle_flag():
    """
    Helper method that imports CourseWaffleFlag from edx-platform at runtime.
    https://github.com/edx/edx-platform/blob/master/openedx/core/djangoapps/waffle_utils/__init__.py#L345
    """
    # pylint: disable=import-error
    from openedx.core.djangoapps.waffle_utils import CourseWaffleFlag
    return CourseWaffleFlag


def import_waffle_flag():
    """
    Helper method that imports WaffleFlag from edx_toggles at runtime.
    """
    # pylint: disable=import-error
    from edx_toggles.toggles import WaffleFlag
    return WaffleFlag


class ConfigMixin:
    """
    Mixin class for determining configuration and feature-toggle state relevant to an ORA block.
    """
    @staticmethod
    def _waffle_switch(switch_name):
        """
        Returns a ``WaffleSwitch`` object in WAFFLE_NAMESPACE
        with the given ``switch_name``.
        """
        # pylint: disable=toggle-missing-annotation
        return WaffleSwitch(f"{WAFFLE_NAMESPACE}.{switch_name}", module_name=__name__)

    @staticmethod
    def _course_waffle_flag(flag_name):
        """
        Returns a ``CourseWaffleFlag`` object in WAFFLE_NAMESPACE
        with the given ``flag_name``.
        """
        CourseWaffleFlag = import_course_waffle_flag()  # pylint: disable=invalid-name
        # pylint: disable=toggle-missing-annotation
        return CourseWaffleFlag(WAFFLE_NAMESPACE, flag_name, module_name=__name__)

    @staticmethod
    def _waffle_flag(flag_name):
        """
        Return a ``WaffleFlag`` object in WAFFLE_NAMESPACE
        with the given ``flag_name``.
        """
        WaffleFlag = import_waffle_flag()  # pylint: disable=invalid-name
        # pylint: disable=toggle-missing-annotation
        return WaffleFlag(f"{WAFFLE_NAMESPACE}.{flag_name}", module_name=__name__)

    @staticmethod
    def _settings_toggle_enabled(toggle_name):
        """
        Returns True iff ``toggle_name`` is defined and set to ``True``
        in Django settings ``FEATURES`` dict.
        """
        if not toggle_name:
            return False

        toggle_state = settings.FEATURES.get(toggle_name)
        # If the feature toggle is not defined in settings, this will return False
        return toggle_state is True

    def is_feature_enabled(self, flag):
        """
        Returns True if a CourseWaffleFlag, WaffleSwitch, or Django settings ``FEATURE``
        is enabled for this block, False otherwise.
        """
        if hasattr(self, 'location') and self._course_waffle_flag(flag).is_enabled(self.location.course_key):
            return True

        if self._waffle_switch(flag).is_enabled():
            return True

        if self._waffle_flag(flag).is_enabled():
            return True

        if self._settings_toggle_enabled(FEATURE_TOGGLES_BY_FLAG_NAME.get(flag)):
            return True

        return False

    @cached_property
    def team_submissions_enabled(self):
        """
        Returns a boolean specifying if the team submission is enabled.
        """
        return self.is_feature_enabled(TEAM_SUBMISSIONS)

    @cached_property
    def user_state_upload_data_enabled(self):
        """
        Returns a boolean indicating the user state upload data flag is enabled or not.
        """
        return self.is_feature_enabled(USER_STATE_UPLOAD_DATA)

    @cached_property
    def is_fetch_all_urls_waffle_enabled(self):
        """
        Returns a boolean indicating the all files urls feature flag is enabled or not.
        """
        return self.is_feature_enabled(ALL_FILES_URLS)

    @cached_property
    def is_mobile_support_enabled(self):
        """
        Returns a boolean indicating if the mobile support feature flag is enabled or not.
        """
        # .. toggle_name: FEATURES['ENABLE_ORA_MOBILE_SUPPORT']
        # .. toggle_implementation: SettingToggle
        # .. toggle_default: False
        # .. toggle_description: Set to True to enable the ORA2 Xblock to be rendered
        #     in mobile apps.
        # .. toggle_use_cases: open_edx
        # .. toggle_creation_date: 2020-10-14
        # .. toggle_tickets: https://github.com/edx/edx-ora2/pull/1445
        return settings.FEATURES.get('ENABLE_ORA_MOBILE_SUPPORT', False)

    @cached_property
    def is_rubric_reuse_enabled(self):
        """
        Return a boolean indicating the reuse of rubric feature is enabled or not.
        """
        # pylint: disable=toggle-missing-target-removal-date
        # .. toggle_name: FEATURES['ENABLE_ORA_RUBRIC_REUSE']
        # .. toggle_implementation: WaffleFlag
        # .. toggle_default: False
        # .. toggle_description: Set to True to enable the reuse of rubric feature
        # .. toggle_use_cases: temporary
        # .. toggle_creation_date: 2021-05-18
        # .. toggle_tickets:  https://openedx.atlassian.net/browse/EDUCATOR-5751
        return self.is_feature_enabled(RUBRIC_REUSE)

    @cached_property
    def is_enhanced_staff_grader_enabled(self):
        """
        Return a boolean indicating the enhanced staff grader feature is enabled or not.
        """
        # .. toggle_name: FEATURES['ENABLE_ENHANCED_STAFF_GRADER']
        # .. toggle_implementation: WaffleFlag
        # .. toggle_default: False
        # .. toggle_description: Set to True to enable the enhanced staff grader feature
        # .. toggle_use_cases: circuit_breaker
        # .. toggle_creation_date: 2021-08-29
        # .. toggle_tickets: https://openedx.atlassian.net/browse/AU-50
        return self.is_feature_enabled(ENHANCED_STAFF_GRADER)



""" File Uploads backends. """


import abc
import mimetypes

from django.conf import settings

from ..exceptions import FileUploadInternalError, FileUploadRequestError


class Settings:
    """Store settings related to file upload

    The following settings are used:

        FILE_UPLOAD_STORAGE_BUCKET_NAME (str, required): name of the bucket
        (AWS or local file directory) to which content will be uploaded.

        FILE_STORAGE_STORAGE_PREFIX (str, defaults to
        DEFAULT_FILE_UPLOAD_STORAGE_PREFIX): this will be used to prefix all
        stored file names. The specified file prefix for the storage must be
        publicly viewable or all uploaded files will not be seen.
    """
    DEFAULT_FILE_UPLOAD_STORAGE_PREFIX = "submissions_attachments"  # pylint: disable=invalid-name
    FILE_EXTENSIONS_BY_TYPE = {
        'image/gif': '.gif',
        'image/jpeg': '.jpg',
        'image/pjpeg': '.jpg',
        'image/png': '.png',
        'application/pdf': '.pdf',
        'application/msword': '.doc',
        'application/vnd.openxmlformats-officedocument.wordprocessingml.document': '.docx'
    }

    @classmethod
    def get_bucket_name(cls):
        bucket_name = getattr(settings, "FILE_UPLOAD_STORAGE_BUCKET_NAME", None)
        if not bucket_name:
            raise FileUploadInternalError("No bucket name configured for FileUpload Service.")
        return bucket_name

    @classmethod
    def get_prefix(cls):
        """Return the prefix for stored files.

        Defaults to the DEFAULT_FILE_UPLOAD_STORAGE_PREFIX class attribute.
        """
        return getattr(settings, "FILE_UPLOAD_STORAGE_PREFIX", cls.DEFAULT_FILE_UPLOAD_STORAGE_PREFIX)

    @classmethod
    def guess_extension(cls, mime_type):
        """
        Guess a file extension (with a leading dot) given its mime type. If no
        file is found, return an empty extension.
        """
        if mime_type in cls.FILE_EXTENSIONS_BY_TYPE:
            return cls.FILE_EXTENSIONS_BY_TYPE[mime_type]
        return mimetypes.guess_extension(mime_type) or ''


class BaseBackend(metaclass=abc.ABCMeta):
    """ Base class for file upload backends. """

    UPLOAD_URL_TIMEOUT = 3600

    # Time (in seconds) before a download url expires
    DOWNLOAD_URL_TIMEOUT = 1000

    @abc.abstractmethod
    def get_upload_url(self, key, content_type):
        """Request a one-time upload URL to upload files.

        Requests a URL for a one-time file upload.

        Args:
            key (str): A unique identifier used to construct the upload location and
                later, can be used to retrieve the same information. This service
                must be able to identify data for both upload and download using
                this key.
            content_type (str): The content type for the file.

        Returns:
            A URL (str) to use for a one-time upload.

        Raises:
            FileUploadInternalError: Raised when an internal error occurs while
                retrieving a one-time URL.
            FileUploadRequestError: Raised when the request failed due to
                request restrictions

        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_download_url(self, key):
        """Requests a URL to download the related file from.

        Requests a URL for the given student_item.

        Args:
            key (str): A unique identifier used to identify the data requested for
                download. This service must be able to identify data for both
                upload and download using this key.

        Returns:
            A URL (str) to use for downloading related files. If no file is found,
            returns an empty string.

        """
        raise NotImplementedError

    @abc.abstractmethod
    def remove_file(self, key):
        """
        Remove file from the storage

        Args:
            key (str): A unique identifier used to identify the data requested for remove.

        Returns:
            True if file was successfully removed or False is file was not removed or was not was not found.
        """
        raise NotImplementedError

    def _retrieve_parameters(self, key):
        """
        Simple utility function to validate settings and arguments before compiling
        bucket names and key names.

        Args:
            key (str): Custom key passed in with the request.

        Returns:
            A tuple of the bucket name and the complete key.

        Raises:
            FileUploadRequestError
            FileUploadInternalError

        """
        if not key:
            raise FileUploadRequestError("Key required for URL request")
        return Settings.get_bucket_name(), self._get_key_name(key)

    def _get_key_name(self, key):
        """Construct a key name with the given string and configured prefix.

        Constructs a unique key with the specified path and the service-specific
        configured prefix.

        Args:
            key (str): Key to identify data for both upload and download.

        Returns:
            A key name (str) to use constructing URLs.
        """
        return "{prefix}/{key}".format(
            prefix=Settings.get_prefix(),
            key=key
        )



"""
Test utilities
"""


from django.core.cache import cache
from django.test import TestCase, TransactionTestCase


def _clear_all_caches():
    """Clear the default cache and any custom caches."""
    cache.clear()


class CacheResetTest(TestCase):
    """
    Test case that resets the cache before and after each test.
    """
    def setUp(self):
        super().setUp()
        _clear_all_caches()

    def tearDown(self):
        super().tearDown()
        _clear_all_caches()


class TransactionCacheResetTest(TransactionTestCase):
    """
    Transaction test case that resets the cache.
    """
    def setUp(self):
        super().setUp()
        _clear_all_caches()

    def tearDown(self):
        super().tearDown()
        _clear_all_caches()



"""
Resolve unspecified dates and date strings to datetimes.
"""


import datetime as dt

from dateutil.parser import parse as parse_date
import pytz


class InvalidDateFormat(Exception):
    """
    The date string could not be parsed.
    """


class DateValidationError(Exception):
    """
    Dates are not semantically valid.
    """


DISTANT_PAST = dt.datetime(dt.MINYEAR, 1, 1, tzinfo=pytz.utc)
DISTANT_FUTURE = dt.datetime(dt.MAXYEAR, 1, 1, tzinfo=pytz.utc)


def _parse_date(value, _):
    """
    Parse an ISO formatted datestring into a datetime object with timezone set to UTC.

    Args:
        value (str or datetime): The ISO formatted date string or datetime object.
        _ (function): The i18n service function used to get the appropriate
            text for a message.

    Returns:
        datetime.datetime

    Raises:
        InvalidDateFormat: The date string could not be parsed.
    """
    if isinstance(value, dt.datetime):
        return value.replace(tzinfo=pytz.utc)

    elif isinstance(value, str):
        try:
            return parse_date(value).replace(tzinfo=pytz.utc)
        except ValueError as ex:
            raise InvalidDateFormat(
                _(
                    "'{date}' is an invalid date format. Make sure the date is formatted as YYYY-MM-DDTHH:MM:SS."
                ).format(date=value)
            ) from ex

    else:
        raise InvalidDateFormat(_("'{date}' must be a date string or datetime").format(date=value))


def parse_date_value(date, _):
    """ Public method for _parse_date """
    return _parse_date(date, _)


def resolve_dates(start, end, date_ranges, _):
    """
    Resolve date strings (including "default" dates) to datetimes.
    The basic rules are:

        1) Unset problem start dates default to the distant past.
        2) Unset problem end dates default to the distant future.
        3) Unset start dates default to the start date of the previous assessment/submission.
            (The first submission defaults to the problem start date.)
        4) Unset end dates default to the end date of the following assessment/submission.
            (The last assessment defaults to the problem end date.)
        5) `start` resolves to the earliest start date.
        6) `end` resolves to the latest end date.
        7) Ensure that `start` is before `end`.
        8) Ensure that `start` is before the earliest due date.
        9) Ensure that `end` is after the latest start date.

    Overriding start/end dates:

        * Rules 5-9 may seem strange, but they're necessary.  Unlike `date_ranges`,
          the `start` and `end` values are inherited by the XBlock from the LMS.
          This means that you can set `start` and `end` in Studio, effectively bypassing
          our validation rules.

        * On the other hand, we *need* the start/due dates so we can resolve unspecified
          date ranges to an actual date.  For example,
          if the problem closes on April 15th, 2014, but the course author hasn't specified
          a due date for a submission, we need ensure the submission closes on April 15th.

        * For this reason, we use `start` and `end` only if they satisfy our validation
          rules.  If not (because a course author has changed them to something invalid in Studio),
          we use the dates that the course author specified in the problem definition,
          which (a) MUST satisfy our ordering constraints, and (b) are probably
          what the author intended.

    Example:

        Suppose I have a problem with a submission and two assessments:

        |                                                                               |
        |    |== submission ==|   |== peer-assessessment ==|   |== self-assessment ==|  |
        |                                                                               |

        and I set start/due dates for the submission and self-assessment, but not for peer-assessment.
        Then by default, peer-assessment will "expand":

        |                                                                               |
        |    |== submission ==|                                |== self-assessment ==|  |
        |    |============================ peer-assessment ==========================|  |
        |                                                                               |

        If I then remove the due date for the submission, but add a due date for peer-assessment:

        |                                                                               |
        |    |== submission =============================|  |== self-assessment ==|     |
        |    |============== peer-assessment ============|                              |
        |                                                                               |

        If no dates are set, start dates default to the distant past and end dates default
        to the distant future:

        |                                                    |
        |    |================= submission ==============|   |
        |    |============== self-assessment ============|   |
        |    |============== peer-assessment ============|   |
        |                                                    |


    Args:
        start (str, ISO date format, or datetime): When the problem opens.
            A value of None indicates that the problem is always open.
        end (str, ISO date format, or datetime): When the problem closes.
            A value of None indicates that the problem never closes.
        date_ranges (list of tuples): list of (start, end) ISO date string tuples indicating
            the start/end timestamps (date string or datetime) of each submission/assessment.
        _ (function): An i18n service function to use for retrieving the
            proper text.

    Returns:
        start (datetime): The resolved start date
        end (datetime): The resolved end date.
        list of (start, end) tuples, where both elements are datetime objects.

    Raises:
        DateValidationError
        InvalidDateFormat
    """
    # Resolve problem start and end dates to minimum and maximum dates
    start = _parse_date(start, _) if start is not None else DISTANT_PAST
    end = _parse_date(end, _) if end is not None else DISTANT_FUTURE
    resolved_starts = []
    resolved_ends = []

    # Amazingly, Studio allows the release date to be after the due date!
    # This can cause a problem if the course author has configured:
    #
    # 1) Problem start >= problem due, and
    # 2) Start/due dates that resolve to the problem start/due date.
    #
    # In this case, all submission/assessment start dates
    # could default to the problem start while
    # due dates default to the problem due date, violating
    # the constraint that start dates always precede due dates.
    # If we detect that the author has done this,
    # we set the start date to just before
    # the due date, so we (just barely) satify the validation rules.
    if start >= end:
        start = end - dt.timedelta(milliseconds=1)

    # Override start/end dates if they fail to satisfy our validation rules
    # These are the only parameters a course author can change in Studio
    # without triggering our validation rules, so we need to use sensible
    # defaults.  See the docstring above for a more detailed justification.
    for step_start, step_end in date_ranges:
        if step_start is not None:
            parsed_start = _parse_date(step_start, _)
            start = min(start, parsed_start)
            end = max(end, parsed_start + dt.timedelta(milliseconds=1))
        if step_end is not None:
            parsed_end = _parse_date(step_end, _)
            end = max(end, parsed_end)
            start = min(start, parsed_end - dt.timedelta(milliseconds=1))

    # Iterate through the list forwards and backwards simultaneously
    # As we iterate forwards, resolve start dates.
    # As we iterate backwards, resolve end dates.
    prev_start = start
    prev_end = end
    for index in range(len(date_ranges)):  # pylint: disable=consider-using-enumerate
        reverse_index = len(date_ranges) - index - 1

        # Resolve "default" start dates to the previous start date.
        # If I set a start date for peer-assessment, but don't set a start date for the following self-assessment,
        # then the self-assessment should default to the same start date as the peer-assessment.
        step_start, __ = date_ranges[index]
        step_start = _parse_date(step_start, _) if step_start is not None else prev_start

        # Resolve "default" end dates to the following end date.
        # If I set a due date for self-assessment, but don't set a due date for the previous peer-assessment,
        # then the peer-assessment should default to the same due date as the self-assessment.
        __, step_end = date_ranges[reverse_index]
        step_end = _parse_date(step_end, _) if step_end is not None else prev_end

        if step_start < prev_start:
            msg = _(
                "This step's start date '{start}' cannot be earlier than the previous step's start date '{prev}'."
            ).format(
                start=step_start,
                prev=prev_start,
            )
            raise DateValidationError(msg)

        if step_end > prev_end:
            msg = _("This step's due date '{due}' cannot be later than the next step's due date '{prev}'.").format(
                due=step_end, prev=prev_end
            )
            raise DateValidationError(msg)

        resolved_starts.append(step_start)
        resolved_ends.insert(0, step_end)
        prev_start = step_start
        prev_end = step_end

    # Combine the resolved dates back into a list of tuples
    resolved_ranges = list(zip(resolved_starts, resolved_ends))

    # Now that we have resolved both start and end dates, we can safely compare them
    for resolved_start, resolved_end in resolved_ranges:
        if resolved_start >= resolved_end:
            msg = _("The start date '{start}' cannot be later than the due date '{due}'").format(
                start=resolved_start, due=resolved_end
            )
            raise DateValidationError(msg)

    return start, end, resolved_ranges



"""
Retrieve user-specific data
"""


def get_user_preferences(user_service):
    """
    Returns the preferred language and timezone for the current user, if specified, or None if not.

    :param user_service: XblockUserService
    """
    user_preferences = {
        'user_language': None,
        'user_timezone': None
    }
    retrieved_preferences = user_service.get_current_user().opt_attrs.get('edx-platform.user_preferences')

    if retrieved_preferences is not None:
        user_preferences['user_timezone'] = retrieved_preferences.get('time_zone')
        user_preferences['user_language'] = retrieved_preferences.get('pref-lang')

    return user_preferences
